"""
Local interpreter for secure Python code execution using Deno/Pyodide.

This module provides PythonInterpreter, which runs Python code in a sandboxed
WASM environment using Deno and Pyodide. It implements the Interpreter
protocol defined in interpreter.py.
"""

import functools
import inspect
import json
import keyword
import logging
import os
import subprocess
import threading
from os import PathLike
from typing import Any, Callable

from dspy.primitives.code_interpreter import SIMPLE_TYPES, CodeInterpreterError, FinalOutput

__all__ = ["PythonInterpreter", "FinalOutput", "CodeInterpreterError"]

logger = logging.getLogger(__name__)

# Pyodide's FFI crashes at exactly 128MB (134,217,728 bytes). Use filesystem
# injection for strings above 100MB to stay safely below this limit.
LARGE_VAR_THRESHOLD = 100 * 1024 * 1024

# =============================================================================
# JSON-RPC 2.0 Helpers
# =============================================================================

# Application errors (range: -32000 to -32099)
JSONRPC_APP_ERRORS = {
    "SyntaxError": -32000,
    "NameError": -32001,
    "TypeError": -32002,
    "ValueError": -32003,
    "AttributeError": -32004,
    "IndexError": -32005,
    "KeyError": -32006,
    "RuntimeError": -32007,
    "CodeInterpreterError": -32008,
    "Unknown": -32099,
}


def _jsonrpc_request(method: str, params: dict, id: int | str) -> str:
    """Create a JSON-RPC 2.0 request (expects response)."""
    return json.dumps({"jsonrpc": "2.0", "method": method, "params": params, "id": id})


def _jsonrpc_notification(method: str, params: dict | None = None) -> str:
    """Create a JSON-RPC 2.0 notification (no response expected)."""
    msg = {"jsonrpc": "2.0", "method": method}
    if params:
        msg["params"] = params
    return json.dumps(msg)


def _jsonrpc_result(result: Any, id: int | str) -> str:
    """Create a JSON-RPC 2.0 success response."""
    return json.dumps({"jsonrpc": "2.0", "result": result, "id": id})


def _jsonrpc_error(code: int, message: str, id: int | str, data: dict | None = None) -> str:
    """Create a JSON-RPC 2.0 error response."""
    err = {"code": code, "message": message}
    if data:
        err["data"] = data
    return json.dumps({"jsonrpc": "2.0", "error": err, "id": id})


class PythonInterpreter:
    """Local interpreter for secure Python execution using Deno and Pyodide.

    Implements the Interpreter protocol for secure code execution in a
    WASM-based sandbox. Code runs in an isolated Pyodide environment with
    no access to the host filesystem, network, or environment by default.

    Prerequisites:
        Deno must be installed: https://docs.deno.com/runtime/getting_started/installation/

    Examples:
        ```python
        # Basic execution
        with PythonInterpreter() as interp:
            result = interp("print(1 + 2)")  # Returns "3"

        # With host-side tools
        def my_tool(question: str) -> str:
            return "answer"

        with PythonInterpreter(tools={"my_tool": my_tool}) as interp:
            result = interp("print(my_tool(question='test'))")
        ```
    """

    def __init__(
        self,
        deno_command: list[str] | None = None,
        enable_read_paths: list[PathLike | str] | None = None,
        enable_write_paths: list[PathLike | str] | None = None,
        enable_env_vars: list[str] | None = None,
        enable_network_access: list[str] | None = None,
        sync_files: bool = True,
        tools: dict[str, Callable[..., str]] | None = None,
        output_fields: list[dict] | None = None,
    ) -> None:
        """
        Args:
            deno_command: command list to launch Deno.
            enable_read_paths: Files or directories to allow reading from in the sandbox.
            enable_write_paths: Files or directories to allow writing to in the sandbox.
                All write paths will also be able to be read from for mounting.
            enable_env_vars: Environment variable names to allow in the sandbox.
            enable_network_access: Domains or IPs to allow network access in the sandbox.
            sync_files: If set, syncs changes within the sandbox back to original files after execution.
            tools: Dictionary mapping tool names to callable functions.
                   Each function should accept keyword arguments and return a string.
                   Tools are callable directly from sandbox code by name.
            output_fields: List of output field definitions for typed SUBMIT signature.
                   Each dict should have 'name' and optionally 'type' keys.
        """
        if isinstance(deno_command, dict):
            raise TypeError("deno_command must be a list of strings, not a dict")

        self.enable_read_paths = enable_read_paths or []
        self.enable_write_paths = enable_write_paths or []
        self.enable_env_vars = enable_env_vars or []
        self.enable_network_access = enable_network_access or []
        self.sync_files = sync_files
        self.tools = dict(tools) if tools else {}
        self.output_fields = output_fields
        self._tools_registered = False
        # TODO later on add enable_run (--allow-run) by proxying subprocess.run through Deno.run() to fix 'emscripten does not support processes' error

        if deno_command:
            self.deno_command = list(deno_command)
        else:
            args = ["deno", "run"]

            # Allow reading runner.js and explicitly enabled paths
            allowed_read_paths = [self._get_runner_path()]

            # Also allow reading Deno's cache directory so Pyodide can load its files
            deno_dir = self._get_deno_dir()
            if deno_dir:
                allowed_read_paths.append(deno_dir)

            if self.enable_read_paths:
                allowed_read_paths.extend(str(p) for p in self.enable_read_paths)
            if self.enable_write_paths:
                allowed_read_paths.extend(str(p) for p in self.enable_write_paths)
            args.append(f"--allow-read={','.join(allowed_read_paths)}")

            self._env_arg = ""
            if self.enable_env_vars:
                user_vars = [str(v).strip() for v in self.enable_env_vars]
                args.append("--allow-env=" + ",".join(user_vars))
                self._env_arg = ",".join(user_vars)
            if self.enable_network_access:
                args.append(f"--allow-net={','.join(str(x) for x in self.enable_network_access)}")
            if self.enable_write_paths:
                args.append(f"--allow-write={','.join(str(x) for x in self.enable_write_paths)}")

            args.append(self._get_runner_path())

            # For runner.js to load in env vars
            if self._env_arg:
                args.append(self._env_arg)
            self.deno_command = args

        self.deno_process = None
        self._mounted_files = False
        self._request_id = 0
        self._owner_thread: int | None = None
        self._pending_large_vars = {}

    def _check_thread_ownership(self) -> None:
        """Ensure this interpreter is only used from a single thread."""
        current_thread = threading.current_thread().ident
        if self._owner_thread is None:
            self._owner_thread = current_thread
        elif self._owner_thread != current_thread:
            raise RuntimeError(
                "PythonInterpreter is not thread-safe and cannot be shared across threads. "
                "Create a separate interpreter instance for each thread."
            )

    @staticmethod
    @functools.lru_cache(maxsize=1)
    def _get_deno_dir() -> str | None:
        if "DENO_DIR" in os.environ:
            return os.environ["DENO_DIR"]

        try:
            result = subprocess.run(
                ["deno", "info", "--json"],
                capture_output=True,
                text=True,
                check=False
            )
            if result.returncode == 0:
                info = json.loads(result.stdout)
                return info.get("denoDir")
        except Exception:
            logger.warning("Unable to find the Deno cache dir.")

        return None

    def _get_runner_path(self) -> str:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(current_dir, "runner.js")

    def _mount_files(self):
        if self._mounted_files:
            return
        paths_to_mount = []
        if self.enable_read_paths:
            paths_to_mount.extend(self.enable_read_paths)
        if self.enable_write_paths:
            paths_to_mount.extend(self.enable_write_paths)
        if not paths_to_mount:
            return
        for path in paths_to_mount:
            if not path:
                continue
            if not os.path.exists(path):
                if self.enable_write_paths and path in self.enable_write_paths:
                    open(path, "a").close()
                else:
                    raise FileNotFoundError(f"Cannot mount non-existent file: {path}")
            virtual_path = f"/sandbox/{os.path.basename(path)}"
            self._send_request("mount_file", {"host_path": str(path), "virtual_path": virtual_path}, f"mounting {path}")
        self._mounted_files = True

    def _sync_files(self):
        if not self.enable_write_paths or not self.sync_files:
            return
        for path in self.enable_write_paths:
            virtual_path = f"/sandbox/{os.path.basename(path)}"
            sync_msg = _jsonrpc_notification("sync_file", {"virtual_path": virtual_path, "host_path": str(path)})
            self.deno_process.stdin.write(sync_msg + "\n")
            self.deno_process.stdin.flush()

    def _extract_parameters(self, fn: Callable) -> list[dict]:
        """Extract parameter info from a callable for sandbox registration."""
        sig = inspect.signature(fn)
        params = []
        for name, param in sig.parameters.items():
            p = {"name": name}
            # Only include type for simple types that work in function signatures
            # Complex types like Union, Optional, etc. are not included
            if param.annotation != inspect.Parameter.empty:
                if param.annotation in SIMPLE_TYPES:
                    p["type"] = param.annotation.__name__
            if param.default != inspect.Parameter.empty:
                p["default"] = param.default
            params.append(p)
        return params

    def _register_tools(self) -> None:
        """Register tools and output fields with the sandbox."""
        if self._tools_registered:
            return

        # Build registration params with typed tool signatures
        params = {}

        if self.tools:
            tools_info = []
            for name, fn in self.tools.items():
                tools_info.append({
                    "name": name,
                    "parameters": self._extract_parameters(fn)
                })
            params["tools"] = tools_info

        if self.output_fields:
            params["outputs"] = self.output_fields

        # Skip if nothing to register
        if not params:
            self._tools_registered = True
            return

        self._send_request("register", params, "registering tools/outputs")
        self._tools_registered = True

    def _handle_tool_call(self, request: dict) -> None:
        """Handle a tool call request from the sandbox."""
        request_id = request["id"]
        params = request.get("params", {})
        tool_name = params.get("name")
        kwargs = params.get("kwargs", {})

        try:
            if tool_name not in self.tools:
                raise CodeInterpreterError(f"Unknown tool: {tool_name}")
            result = self.tools[tool_name](**kwargs)
            is_json = isinstance(result, (list, dict))
            response = _jsonrpc_result(
                {"value": json.dumps(result) if is_json else (str(result) if result is not None else ""), "type": "json" if is_json else "string"},
                request_id
            )
        except Exception as e:
            error_type = type(e).__name__
            error_code = JSONRPC_APP_ERRORS.get(error_type, JSONRPC_APP_ERRORS["Unknown"])
            response = _jsonrpc_error(error_code, str(e), request_id, {"type": error_type})

        self.deno_process.stdin.write(response + "\n")
        self.deno_process.stdin.flush()

    def _ensure_deno_process(self) -> None:
        if self.deno_process is None or self.deno_process.poll() is not None:
            # Process identity changed (or process missing), so replay setup.
            self._tools_registered = False
            self._mounted_files = False
            try:
                self.deno_process = subprocess.Popen(
                    self.deno_command,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                    encoding="UTF-8",
                    env=os.environ.copy()
                )
            except FileNotFoundError as e:
                install_instructions = (
                    "Deno executable not found. Please install Deno to proceed.\n"
                    "Installation instructions:\n"
                    "> curl -fsSL https://deno.land/install.sh | sh\n"
                    "*or*, on macOS with Homebrew:\n"
                    "> brew install deno\n"
                    "For additional configurations: https://docs.deno.com/runtime/getting_started/installation/"
                )
                raise CodeInterpreterError(install_instructions) from e
            self._health_check()

    _MAX_SKIP_LINES = 100

    def _read_response_line(self, context: str) -> str:
        """Read one stdout line from Deno or raise a process-level error."""
        response_line = self.deno_process.stdout.readline().strip()
        if response_line:
            return response_line

        exit_code = self.deno_process.poll()
        if exit_code is not None:
            stderr = self.deno_process.stderr.read() if self.deno_process.stderr else ""
            raise CodeInterpreterError(f"Deno exited (code {exit_code}) {context}: {stderr}")
        raise CodeInterpreterError(f"No response {context}")

    def _parse_response_line(self, response_line: str, context: str) -> dict | None:
        """Parse a JSON-RPC line, returning None for non-JSON or malformed lines."""
        if not response_line.startswith("{"):
            logger.debug("Skipping non-JSON output during %s: %s", context, response_line)
            return None

        try:
            return json.loads(response_line)
        except json.JSONDecodeError:
            logger.debug("Skipping malformed JSON during %s: %s", context, response_line[:100])
            return None

    def _send_request(self, method: str, params: dict, context: str) -> dict:
        """Send a JSON-RPC request and return the parsed response.

        Non-JSON lines (e.g. Pyodide package loading messages) are skipped,
        up to ``_MAX_SKIP_LINES`` to prevent unbounded blocking.
        """
        self._request_id += 1
        request_id = self._request_id
        msg = _jsonrpc_request(method, params, request_id)
        self.deno_process.stdin.write(msg + "\n")
        self.deno_process.stdin.flush()

        skipped = 0
        while skipped <= self._MAX_SKIP_LINES:
            response_line = self._read_response_line(context)
            response = self._parse_response_line(response_line, context)
            if response is None:
                skipped += 1
                continue

            if response.get("id") != request_id:
                raise CodeInterpreterError(f"Response ID mismatch {context}: expected {request_id}, got {response.get('id')}")
            if "error" in response:
                raise CodeInterpreterError(f"Error {context}: {response['error'].get('message', 'Unknown error')}")
            return response

        raise CodeInterpreterError(f"Too many non-JSON lines ({skipped}) {context}")

    def _health_check(self) -> None:
        """Verify the subprocess is alive by executing a simple expression."""
        response = self._send_request("execute", {"code": "print(1+1)"}, "during health check")
        if response.get("result", {}).get("output", "").strip() != "2":
            raise CodeInterpreterError(f"Unexpected ping response: {response}")

    def _to_json_compatible(self, value: Any) -> Any:
        """Recursively convert Python values to JSON-compatible types."""
        if value is None or isinstance(value, (str, int, float, bool)):
            return value
        elif isinstance(value, dict):
            return {k: self._to_json_compatible(v) for k, v in value.items()}
        elif isinstance(value, (list, tuple)):
            return [self._to_json_compatible(v) for v in value]
        elif isinstance(value, set):
            try:
                return sorted(self._to_json_compatible(v) for v in value)
            except TypeError:
                return [self._to_json_compatible(v) for v in value]
        else:
            raise CodeInterpreterError(f"Unsupported value type: {type(value).__name__}")

    def _inject_variables(self, code: str, variables: dict[str, Any]) -> str:
        """Insert Python assignments for each variable at the top of the code."""
        for key in variables:
            if not key.isidentifier() or keyword.iskeyword(key) or key == "json":
                raise CodeInterpreterError(f"Invalid variable name: '{key}'")

        large_vars = {}
        small_assignments = []
        for k, v in variables.items():
            serialized = self._serialize_value(v)
            if len(serialized) > LARGE_VAR_THRESHOLD:
                large_vars[k] = json.dumps(self._to_json_compatible(v))
            else:
                small_assignments.append(f"{k} = {serialized}")

        self._pending_large_vars = large_vars

        if large_vars:
            large_assignments = [f"{k} = json.loads(open('/tmp/dspy_vars/{k}.json').read())" for k in large_vars]
            assignments = ["import json"] + small_assignments + large_assignments
        else:
            assignments = small_assignments

        return "\n".join(assignments) + "\n" + code if assignments else code

    def _serialize_value(self, value: Any) -> str:
        """Serialize a Python value to a Python literal string for injection.

        Sets and tuples are converted to lists for JSON round-trip compatibility,
        since the sandbox returns values via JSON which doesn't support these types.
        """
        if value is None:
            return "None"
        elif isinstance(value, str):
            return repr(value)
        elif isinstance(value, bool):
            # Must check bool before int since bool is a subclass of int
            return "True" if value else "False"
        elif isinstance(value, (int, float)):
            return str(value)
        elif isinstance(value, (list, tuple)):
            # Tuples become lists for JSON compatibility
            items = ", ".join(self._serialize_value(item) for item in value)
            return f"[{items}]"
        elif isinstance(value, dict):
            items = ", ".join(
                f"{self._serialize_value(k)}: {self._serialize_value(v)}"
                for k, v in value.items()
            )
            return f"{{{items}}}"
        elif isinstance(value, set):
            # Sets become sorted lists (or unsorted if mixed types) for JSON compatibility
            try:
                sorted_items = sorted(value)
            except TypeError:
                sorted_items = list(value)
            items = ", ".join(self._serialize_value(item) for item in sorted_items)
            return f"[{items}]"
        else:
            raise CodeInterpreterError(f"Unsupported value type: {type(value).__name__}")

    def _inject_large_var(self, name: str, value: str) -> None:
        """Inject a large variable via the virtual filesystem."""
        self._send_request("inject_var", {"name": name, "value": value}, f"injecting variable '{name}'")

    def execute(
        self,
        code: str,
        variables: dict[str, Any] | None = None,
    ) -> Any:
        self._check_thread_ownership()
        variables = variables or {}
        code = self._inject_variables(code, variables)
        self._ensure_deno_process()
        self._mount_files()
        self._register_tools()

        for name, value in self._pending_large_vars.items():
            self._inject_large_var(name, value)

        # Send the code as JSON-RPC request
        self._request_id += 1
        execute_request_id = self._request_id
        input_data = _jsonrpc_request("execute", {"code": code}, execute_request_id)
        try:
            self.deno_process.stdin.write(input_data + "\n")
            self.deno_process.stdin.flush()
        except BrokenPipeError:
            # If the process died, restart and try again once
            self._ensure_deno_process()
            self._mount_files()
            self._register_tools()
            for name, value in self._pending_large_vars.items():
                self._inject_large_var(name, value)
            self.deno_process.stdin.write(input_data + "\n")
            self.deno_process.stdin.flush()

        # Read and handle messages until we get the final output.
        # Loop is needed because tool calls require back-and-forth communication.
        skipped = 0
        while skipped <= self._MAX_SKIP_LINES:
            output_line = self._read_response_line("during execution")
            msg = self._parse_response_line(output_line, "during execution")
            if msg is None:
                skipped += 1
                continue

            # Handle incoming requests (tool calls from sandbox)
            if "method" in msg:
                if msg["method"] == "tool_call":
                    self._handle_tool_call(msg)
                    continue

            # Handle success response
            if "result" in msg:
                if msg.get("id") != execute_request_id:
                    raise CodeInterpreterError(f"Response ID mismatch: expected {execute_request_id}, got {msg.get('id')}")
                result = msg["result"]
                self._sync_files()
                # Check for SUBMIT (encoded as success with "final" field)
                if "final" in result:
                    return FinalOutput(result["final"])
                return result.get("output", None)

            # Handle error response
            if "error" in msg:
                # Errors with id=null are unsolicited errors (e.g., unhandled async rejections)
                # Treat them as errors for the current request
                if msg.get("id") is not None and msg.get("id") != execute_request_id:
                    raise CodeInterpreterError(f"Response ID mismatch: expected {execute_request_id}, got {msg.get('id')}")
                error = msg["error"]
                error_code = error.get("code", JSONRPC_APP_ERRORS["Unknown"])
                error_message = error.get("message", "Unknown error")
                error_data = error.get("data", {})
                error_type = error_data.get("type", "Error")

                if error_code == JSONRPC_APP_ERRORS["SyntaxError"]:
                    raise SyntaxError(f"Invalid Python syntax. message: {error_message}")
                else:
                    raise CodeInterpreterError(f"{error_type}: {error_data.get('args') or error_message}")

            # Unexpected message format - neither a recognized method nor a response
            raise CodeInterpreterError(f"Unexpected message format from sandbox: {msg}")

        raise CodeInterpreterError(f"Too many non-JSON lines ({skipped}) during execution")

    def start(self) -> None:
        """Initialize the Deno/Pyodide sandbox.

        This pre-warms the sandbox by starting the Deno subprocess.
        Can be called explicitly for pooling, or will be called lazily
        on first execute().

        Idempotent: safe to call multiple times.
        """
        self._ensure_deno_process()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.shutdown()

    def __call__(
        self,
        code: str,
        variables: dict[str, Any] | None = None,
    ) -> Any:
        return self.execute(code, variables)

    def shutdown(self) -> None:
        if self.deno_process and self.deno_process.poll() is None:
            self.deno_process.stdin.write(_jsonrpc_notification("shutdown") + "\n")
            self.deno_process.stdin.flush()
            self.deno_process.stdin.close()
            self.deno_process.wait()
        self.deno_process = None
        self._owner_thread = None



"""
REPL data types for RLM and interpreter interactions.

These types represent the state and history of REPL-based execution:
- REPLVariable: Metadata about variables available in the REPL
- REPLEntry: A single interaction (reasoning, code, output)
- REPLHistory: Container for the full interaction history
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Iterator

import pydantic
from pydantic import Field

from dspy.adapters.utils import serialize_for_json

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo

__all__ = ["REPLVariable", "REPLEntry", "REPLHistory"]


class REPLVariable(pydantic.BaseModel):
    """Metadata about a variable available in the REPL environment."""

    name: str
    type_name: str
    desc: str = ""
    constraints: str = ""
    total_length: int
    preview: str

    model_config = pydantic.ConfigDict(frozen=True)

    @classmethod
    def from_value(
        cls,
        name: str,
        value: Any,
        field_info: FieldInfo | None = None,
        preview_chars: int = 1000,
    ) -> REPLVariable:
        """Create REPLVariable from an actual value and optional field info.

        Args:
            name: Variable name
            value: The actual value
            field_info: Optional pydantic FieldInfo with desc/constraints metadata
            preview_chars: Max characters for preview
        """
        jsonable = serialize_for_json(value)
        if isinstance(jsonable, (dict, list)):
            value_str = json.dumps(jsonable, indent=2)
        else:
            value_str = str(jsonable)
        is_truncated = len(value_str) > preview_chars
        if is_truncated:
            half = preview_chars // 2
            preview = value_str[:half] + "..." + value_str[-half:]
        else:
            preview = value_str

        # Extract desc and constraints from field_info if provided
        desc = ""
        constraints = ""
        if field_info and hasattr(field_info, "json_schema_extra") and field_info.json_schema_extra:
            raw_desc = field_info.json_schema_extra.get("desc", "")
            # Skip placeholder descs like "${name}"
            if raw_desc and not raw_desc.startswith("${"):
                desc = raw_desc
            constraints = field_info.json_schema_extra.get("constraints", "")

        return cls(
            name=name,
            type_name=type(value).__name__,
            desc=desc,
            constraints=constraints,
            total_length=len(value_str),
            preview=preview,
        )

    def format(self) -> str:
        """Format variable metadata for prompt inclusion."""
        lines = [f"Variable: `{self.name}` (access it in your code)"]
        lines.append(f"Type: {self.type_name}")
        if self.desc:
            lines.append(f"Description: {self.desc}")
        if self.constraints:
            lines.append(f"Constraints: {self.constraints}")
        lines.append(f"Total length: {self.total_length:,} characters")
        lines.append(f"Preview:\n```\n{self.preview}\n```")
        return "\n".join(lines)

    @pydantic.model_serializer()
    def serialize_model(self) -> str:
        return self.format()


class REPLEntry(pydantic.BaseModel):
    """A single REPL interaction entry containing reasoning, code, and output."""

    reasoning: str = ""
    code: str
    output: str

    model_config = pydantic.ConfigDict(frozen=True)

    @staticmethod
    def format_output(output: str, max_output_chars: int = 10_000) -> str:
        """Format output with head+tail truncation, preserving true length in header."""
        raw_len = len(output)
        if raw_len > max_output_chars:
            half = max_output_chars // 2
            omitted = raw_len - max_output_chars
            output = output[:half] + f"\n\n... ({omitted:,} characters omitted) ...\n\n" + output[-half:]
        return f"Output ({raw_len:,} chars):\n{output}"

    def format(self, index: int, max_output_chars: int = 10_000) -> str:
        """Format this entry for inclusion in prompts."""
        reasoning_line = f"Reasoning: {self.reasoning}\n" if self.reasoning else ""
        code_block = f"```python\n{self.code}\n```"
        return f"=== Step {index + 1} ===\n{reasoning_line}Code:\n{code_block}\n{self.format_output(self.output, max_output_chars)}"


class REPLHistory(pydantic.BaseModel):
    """Container for REPL interaction history.

    Immutable: append() returns a new instance with the entry added.
    """

    entries: list[REPLEntry] = Field(default_factory=list)
    max_output_chars: int = 10_000

    model_config = pydantic.ConfigDict(frozen=True)

    def format(self) -> str:
        if not self.entries:
            return "You have not interacted with the REPL environment yet."
        return "\n".join(entry.format(index=i, max_output_chars=self.max_output_chars) for i, entry in enumerate(self.entries))

    @pydantic.model_serializer()
    def serialize_model(self) -> str:
        return self.format()

    def append(self, *, reasoning: str = "", code: str, output: str) -> REPLHistory:
        """Return a new REPLHistory with the entry appended."""
        new_entry = REPLEntry(reasoning=reasoning, code=code, output=output)
        return REPLHistory(entries=list(self.entries) + [new_entry], max_output_chars=self.max_output_chars)

    def __len__(self) -> int:
        return len(self.entries)

    def __iter__(self) -> Iterator[REPLEntry]:
        return iter(self.entries)

    def __bool__(self) -> bool:
        return len(self.entries) > 0



"""
Custom adapter for improving structured outputs using the information from Pydantic models.
Based on the format used by BAML: https://github.com/BoundaryML/baml
"""

import inspect
import types
from typing import Any, Literal, Union, get_args, get_origin

from pydantic import BaseModel

from dspy.adapters.json_adapter import JSONAdapter
from dspy.adapters.utils import format_field_value as original_format_field_value
from dspy.signatures.signature import Signature

# Changing the comment symbol to Python's # rather than other languages' // seems to help
COMMENT_SYMBOL = "#"
INDENTATION = "  "


def _render_type_str(
    annotation: Any,
    depth: int = 0,
    indent: int = 0,
    seen_models: set[type] | None = None,
) -> str:
    """Recursively renders a type annotation into a simplified string.

    Args:
        annotation: The type annotation to render
        depth: Current recursion depth (prevents infinite recursion)
        indent: Current indentation level for nested structures
    """
    # Non-nested types
    if annotation is str:
        return "string"
    if annotation is int:
        return "int"
    if annotation is float:
        return "float"
    if annotation is bool:
        return "boolean"
    if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
        return _build_simplified_schema(annotation, indent, seen_models)

    try:
        origin = get_origin(annotation)
        args = get_args(annotation)
    except Exception:
        return str(annotation)

    # Optional[T] or T | None
    if origin in (types.UnionType, Union):
        non_none_args = [arg for arg in args if arg is not type(None)]
        # Render the non-None part of the union
        type_render = " or ".join([_render_type_str(arg, depth + 1, indent, seen_models) for arg in non_none_args])
        # Add "or null" if None was part of the union
        if len(non_none_args) < len(args):
            return f"{type_render} or null"
        return type_render

    # Literal[T1, T2, ...]
    if origin is Literal:
        return " or ".join(f'"{arg}"' for arg in args)

    # list[T]
    if origin is list:
        # For Pydantic models in lists, use bracket notation
        inner_type = args[0]
        if inspect.isclass(inner_type) and issubclass(inner_type, BaseModel):
            # Build inner schema - the Pydantic model inside should use indent level for array contents
            inner_schema = _build_simplified_schema(inner_type, indent + 1, seen_models)
            # Format with proper bracket notation and indentation
            current_indent = INDENTATION * indent
            return f"[\n{inner_schema}\n{current_indent}]"
        else:
            return f"{_render_type_str(inner_type, depth + 1, indent, seen_models)}[]"

    # dict[T1, T2]
    if origin is dict:
        return f"dict[{_render_type_str(args[0], depth + 1, indent, seen_models)}, {_render_type_str(args[1], depth + 1, indent, seen_models)}]"

    # fallback
    if hasattr(annotation, "__name__"):
        return annotation.__name__
    return str(annotation)


def _build_simplified_schema(
    pydantic_model: type[BaseModel],
    indent: int = 0,
    seen_models: set[type] | None = None,
) -> str:
    """Builds a simplified, human-readable schema from a Pydantic model.

    Args:
        pydantic_model: The Pydantic model to build schema for
        indent: Current indentation level
        seen_models: Set to track visited pydantic models (prevents infinite recursion)
    """
    seen_models = seen_models or set()

    if pydantic_model in seen_models:
        raise ValueError("BAMLAdapter cannot handle recursive pydantic models, please use a different adapter.")

    # Add `pydantic_model` to `seen_models` with a placeholder value to avoid infinite recursion.
    seen_models.add(pydantic_model)

    lines = []
    current_indent = INDENTATION * indent
    next_indent = INDENTATION * (indent + 1)

    # Add model docstring as a comment above the object if it exists
    # Only do this for top-level schemas (indent=0), since nested field docstrings
    # are already added before the field name in the parent schema
    if indent == 0 and pydantic_model.__doc__:
        docstring = pydantic_model.__doc__.strip()
        # Handle multiline docstrings by prefixing each line with the comment symbol
        for line in docstring.split("\n"):
            line = line.strip()
            if line:
                lines.append(f"{current_indent}{COMMENT_SYMBOL} {line}")

    lines.append(f"{current_indent}{{")

    fields = pydantic_model.model_fields
    if not fields:
        lines.append(f"{next_indent}{COMMENT_SYMBOL} No fields defined")
    for name, field in fields.items():
        if field.description:
            lines.append(f"{next_indent}{COMMENT_SYMBOL} {field.description}")
        elif field.alias and field.alias != name:
            # If there's an alias but no description, show the alias as a comment
            lines.append(f"{next_indent}{COMMENT_SYMBOL} alias: {field.alias}")

        # If the field type is a BaseModel, add its docstring as a comment before the field
        field_annotation = field.annotation
        # Handle Optional types
        origin = get_origin(field_annotation)
        if origin in (types.UnionType, Union):
            args = get_args(field_annotation)
            non_none_args = [arg for arg in args if arg is not type(None)]
            if len(non_none_args) == 1:
                field_annotation = non_none_args[0]

        if inspect.isclass(field_annotation) and issubclass(field_annotation, BaseModel):
            if field_annotation.__doc__:
                docstring = field_annotation.__doc__.strip()
                for line in docstring.split("\n"):
                    line = line.strip()
                    if line:
                        lines.append(f"{next_indent}{COMMENT_SYMBOL} {line}")

        rendered_type = _render_type_str(field.annotation, indent=indent + 1, seen_models=seen_models)
        line = f"{next_indent}{name}: {rendered_type},"

        lines.append(line)

    lines.append(f"{current_indent}}}")
    return "\n".join(lines)


class BAMLAdapter(JSONAdapter):
    """
    A DSPy adapter that improves the rendering of complex/nested Pydantic models to help LMs.

    This adapter generates a compact, human-readable schema representation for nested Pydantic output
    fields, inspired by the BAML project's JSON formatter (https://github.com/BoundaryML/baml).
    The resulting rendered schema is more token-efficient and easier for smaller LMs to follow than a
    raw JSON schema. It also includes Pydantic field descriptions as comments in the schema, which
    provide valuable additional context for the LM to understand the expected output.

    Example Usage:
    ```python
    import dspy
    from pydantic import BaseModel, Field
    from typing import Literal
    from baml_adapter import BAMLAdapter  # Import from your module

    # 1. Define your Pydantic models
    class PatientAddress(BaseModel):
        street: str
        city: str
        country: Literal["US", "CA"]

    class PatientDetails(BaseModel):
        name: str = Field(description="Full name of the patient.")
        age: int
        address: PatientAddress | None

    # 2. Define a signature using the Pydantic model as an output field
    class ExtractPatientInfo(dspy.Signature):
        '''Extract patient information from the clinical note.'''
        clinical_note: str = dspy.InputField()
        patient_info: PatientDetails = dspy.OutputField()

    # 3. Configure dspy to use the new adapter
    llm = dspy.OpenAI(model="gpt-4.1-mini")
    dspy.configure(lm=llm, adapter=BAMLAdapter())

    # 4. Run your program
    extractor = dspy.Predict(ExtractPatientInfo)
    note = "John Doe, 45 years old, lives at 123 Main St, Anytown. Resident of the US."
    result = extractor(clinical_note=note)
    print(result.patient_info)

    # Expected output:
    # PatientDetails(name='John Doe', age=45, address=PatientAddress(street='123 Main St', city='Anytown', country='US'))
    ```
    """

    def format_field_structure(self, signature: type[Signature]) -> str:
        """Overrides the base method to generate a simplified schema for Pydantic models."""

        sections = []

        # Add structural explanation
        sections.append(
            "All interactions will be structured in the following way, with the appropriate values filled in.\n"
        )

        # Add input structure section
        if signature.input_fields:
            for name in signature.input_fields.keys():
                sections.append(f"[[ ## {name} ## ]]")
                sections.append(f"{{{name}}}")
                sections.append("")  # Empty line after each input

        # Add output structure section
        if signature.output_fields:
            for name, field in signature.output_fields.items():
                field_type = field.annotation
                sections.append(f"[[ ## {name} ## ]]")
                sections.append(f"Output field `{name}` should be of type: {_render_type_str(field_type, indent=0)}\n")

        # Add completed section
        sections.append("[[ ## completed ## ]]")

        return "\n".join(sections)

    def format_user_message_content(
        self,
        signature: type[Signature],
        inputs: dict[str, Any],
        prefix: str = "",
        suffix: str = "",
        main_request: bool = False,
    ) -> str:
        """Overrides the base method to render Pydantic input instances as clean JSON."""
        messages = [prefix]
        for key, field_info in signature.input_fields.items():
            if key in inputs:
                value = inputs.get(key)
                formatted_value = ""
                if isinstance(value, BaseModel):
                    # Use clean, indented JSON for Pydantic instances
                    formatted_value = value.model_dump_json(indent=2, by_alias=True)
                else:
                    # Fallback to the original dspy formatter for other types
                    formatted_value = original_format_field_value(field_info=field_info, value=value)

                messages.append(f"[[ ## {key} ## ]]\n{formatted_value}")

        if main_request:
            output_requirements = self.user_message_output_requirements(signature)
            if output_requirements is not None:
                messages.append(output_requirements)

        messages.append(suffix)
        return "\n\n".join(m for m in messages if m).strip()



import base64
import io
import mimetypes
import os
import warnings
from functools import lru_cache
from typing import Any, Union
from urllib.parse import urlparse

import pydantic
import requests

from dspy.adapters.types.base_type import Type

try:
    from PIL import Image as PILImage

    PIL_AVAILABLE = True
except ImportError:
    PIL_AVAILABLE = False


class Image(Type):
    url: str

    model_config = pydantic.ConfigDict(
        frozen=True,
        str_strip_whitespace=True,
        validate_assignment=True,
        extra="forbid",
    )

    def __init__(self, url: Any = None, *, download: bool = False, verify: bool = True, **data):
        """Create an Image.

        Parameters
        ----------
        url:
            The image source. Supported values include

            - ``str``: HTTP(S)/GS URL or local file path
            - ``bytes``: raw image bytes
            - ``PIL.Image.Image``: a PIL image instance
            - ``dict`` with a single ``{"url": value}`` entry (legacy form)
            - already encoded data URI

        download:
            Whether remote URLs should be downloaded to infer their MIME type.

        verify:
            Whether to verify SSL certificates when downloading images from URLs.
            Set to False for self-signed certificates. Default is True.

        Any additional keyword arguments are passed to :class:`pydantic.BaseModel`.
        """

        if url is not None and "url" not in data:
            # Support a positional argument while allowing ``url=`` in **data.
            if isinstance(url, dict) and set(url.keys()) == {"url"}:
                # Legacy dict form from previous model validator.
                data["url"] = url["url"]
            else:
                # ``url`` may be a string, bytes, or a PIL image.
                data["url"] = url

        if "url" in data:
            # Normalize any accepted input into a base64 data URI or plain URL.
            data["url"] = encode_image(data["url"], download_images=download, verify=verify)

        # Delegate the rest of initialization to pydantic's BaseModel.
        super().__init__(**data)

    @lru_cache(maxsize=32)
    def format(self) -> list[dict[str, Any]] | str:
        try:
            image_url = encode_image(self.url)
        except Exception as e:
            raise ValueError(f"Failed to format image for DSPy: {e}")
        return [{"type": "image_url", "image_url": {"url": image_url}}]

    @classmethod
    def from_url(cls, url: str, download: bool = False):
        warnings.warn(
            "Image.from_url is deprecated; use Image(url) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls(url, download=download)

    @classmethod
    def from_file(cls, file_path: str):
        warnings.warn(
            "Image.from_file is deprecated; use Image(file_path) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls(file_path)

    @classmethod
    def from_PIL(cls, pil_image):  # noqa: N802
        warnings.warn(
            "Image.from_PIL is deprecated; use Image(pil_image) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls(pil_image)

    def __str__(self):
        return self.serialize_model()

    def __repr__(self):
        if "base64" in self.url:
            len_base64 = len(self.url.split("base64,")[1])
            image_type = self.url.split(";")[0].split("/")[-1]
            return f"Image(url=data:image/{image_type};base64,<IMAGE_BASE_64_ENCODED({len_base64!s})>)"
        return f"Image(url='{self.url}')"


def is_url(string: str) -> bool:
    """Check if a string is a valid URL."""
    try:
        result = urlparse(string)
        return all([result.scheme in ("http", "https", "gs"), result.netloc])
    except ValueError:
        return False


def encode_image(image: Union[str, bytes, "PILImage.Image", dict], download_images: bool = False, verify: bool = True) -> str:
    """
    Encode an image or file to a base64 data URI.

    Args:
        image: The image or file to encode. Can be a PIL Image, file path, URL, or data URI.
        download_images: Whether to download images from URLs.
        verify: Whether to verify SSL certificates when downloading images.

    Returns:
        str: The data URI of the file or the URL if download_images is False.

    Raises:
        ValueError: If the file type is not supported.
    """
    if isinstance(image, dict) and "url" in image:
        # NOTE: Not doing other validation for now
        return image["url"]
    elif isinstance(image, str):
        if image.startswith("data:"):
            # Already a data URI
            return image
        elif os.path.isfile(image):
            # File path
            return _encode_image_from_file(image)
        elif is_url(image):
            # URL
            if download_images:
                return _encode_image_from_url(image, verify=verify)
            else:
                # Return the URL as is
                return image
        else:
            # Unsupported string format
            raise ValueError(f"Unrecognized file string: {image}; If this file type should be supported, please open an issue.")
    elif PIL_AVAILABLE and isinstance(image, PILImage.Image):
        # PIL Image
        return _encode_pil_image(image)
    elif isinstance(image, bytes):
        # Raw bytes
        if not PIL_AVAILABLE:
            raise ImportError("Pillow is required to process image bytes.")
        img = PILImage.open(io.BytesIO(image))
        return _encode_pil_image(img)
    elif isinstance(image, Image):
        return image.url
    else:
        print(f"Unsupported image type: {type(image)}")
        raise ValueError(f"Unsupported image type: {type(image)}")


def _encode_image_from_file(file_path: str) -> str:
    """Encode a file from a file path to a base64 data URI."""
    with open(file_path, "rb") as file:
        file_data = file.read()

    # Use mimetypes to guess directly from the file path
    mime_type, _ = mimetypes.guess_type(file_path)
    if mime_type is None:
        raise ValueError(f"Could not determine MIME type for file: {file_path}")

    encoded_data = base64.b64encode(file_data).decode("utf-8")
    return f"data:{mime_type};base64,{encoded_data}"


def _encode_image_from_url(image_url: str, verify: bool = True) -> str:
    """Encode a file from a URL to a base64 data URI.
    
    Args:
        image_url: The URL of the image to download.
        verify: Whether to verify SSL certificates. Set to False for self-signed certs.
    """
    response = requests.get(image_url, verify=verify)
    response.raise_for_status()
    content_type = response.headers.get("Content-Type", "")

    # Use the content type from the response headers if available
    if content_type:
        mime_type = content_type
    else:
        # Try to guess MIME type from URL
        mime_type, _ = mimetypes.guess_type(image_url)
        if mime_type is None:
            raise ValueError(f"Could not determine MIME type for URL: {image_url}")

    encoded_data = base64.b64encode(response.content).decode("utf-8")
    return f"data:{mime_type};base64,{encoded_data}"


def _encode_pil_image(image: "PILImage") -> str:
    """Encode a PIL Image object to a base64 data URI."""
    buffered = io.BytesIO()
    file_format = image.format or "PNG"
    image.save(buffered, format=file_format)

    # Get the correct MIME type using the image format
    file_extension = file_format.lower()
    mime_type, _ = mimetypes.guess_type(f"file.{file_extension}")
    if mime_type is None:
        raise ValueError(f"Could not determine MIME type for image format: {file_format}")

    encoded_data = base64.b64encode(buffered.getvalue()).decode("utf-8")
    return f"data:{mime_type};base64,{encoded_data}"


def _get_file_extension(path_or_url: str) -> str:
    """Extract the file extension from a file path or URL."""
    extension = os.path.splitext(urlparse(path_or_url).path)[1].lstrip(".").lower()
    return extension or "png"  # Default to 'png' if no extension found


def is_image(obj) -> bool:
    """Check if the object is an image or a valid media file reference."""
    if PIL_AVAILABLE and isinstance(obj, PILImage.Image):
        return True
    if isinstance(obj, str):
        if obj.startswith("data:"):
            return True
        elif os.path.isfile(obj):
            return True
        elif is_url(obj):
            return True
    return False



import asyncio
import contextvars
import copy
import logging
import threading
from contextlib import contextmanager
from typing import Any

import cloudpickle

from dspy.dsp.utils.utils import dotdict

logger = logging.getLogger(__name__)

DEFAULT_CONFIG = dotdict(
    lm=None,
    adapter=None,
    rm=None,
    branch_idx=0,
    trace=[],
    callbacks=[],
    async_max_workers=8,
    send_stream=None,
    disable_history=False,
    track_usage=False,
    usage_tracker=None,
    caller_predict=None,
    caller_modules=None,
    stream_listeners=[],
    provide_traceback=False,  # Whether to include traceback information in error logs.
    num_threads=8,  # Number of threads to use for parallel processing.
    max_errors=10,  # Maximum errors before halting operations.
    # If true, async tools can be called in sync mode by getting converted to sync.
    allow_tool_async_sync_conversion=False,
    max_history_size=10000,
    max_trace_size=10000,
    warn_on_type_mismatch=True,  # Whether to log warnings when a module's input type doesn't match the signature type.
)

# Global base configuration and owner tracking
main_thread_config = copy.deepcopy(DEFAULT_CONFIG)
config_owner_thread_id = None
config_owner_async_task = None

# Global lock for settings configuration
global_lock = threading.Lock()

thread_local_overrides = contextvars.ContextVar("context_overrides", default=dotdict())


class Settings:
    """
    A singleton class for DSPy configuration settings.
    Thread-safe global configuration.
    - 'configure' can be called by only one 'owner' thread (the first thread that calls it).
    - Other threads see the configured global values from 'main_thread_config'.
    - 'context' sets thread-local overrides. These overrides propagate to threads spawned
      inside that context block, when (and only when!) using a ParallelExecutor that copies overrides.

      1. Only one unique thread (which can be any thread!) can call dspy.configure.
      2. It affects a global state, visible to all. As a result, user threads work, but they shouldn't be
         mixed with concurrent changes to dspy.configure from the "main" thread.
         (TODO: In the future, add warnings: if there are near-in-time user-thread reads followed by .configure calls.)
      3. Any thread can use dspy.context. It propagates to child threads created with DSPy primitives: Parallel, asyncify, etc.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @property
    def lock(self):
        return global_lock

    def __getattr__(self, name):
        overrides = thread_local_overrides.get()
        if name in overrides:
            return overrides[name]
        elif name in main_thread_config:
            return main_thread_config[name]
        else:
            raise AttributeError(f"'Settings' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name in ("_instance",):
            super().__setattr__(name, value)
        else:
            self.configure(**{name: value})

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __contains__(self, key):
        overrides = thread_local_overrides.get()
        return key in overrides or key in main_thread_config

    def get(self, key, default=None):
        try:
            return self[key]
        except AttributeError:
            return default

    def copy(self):
        overrides = thread_local_overrides.get()
        return dotdict({**main_thread_config, **overrides})

    @property
    def config(self):
        return self.copy()

    def _ensure_configure_allowed(self):
        global main_thread_config, config_owner_thread_id, config_owner_async_task
        current_thread_id = threading.get_ident()

        if config_owner_thread_id is None:
            # First `configure` call assigns the owner thread id.
            config_owner_thread_id = current_thread_id

        if config_owner_thread_id != current_thread_id:
            # Disallow a second `configure` calls from other threads.
            raise RuntimeError("dspy.settings can only be changed by the thread that initially configured it.")

        # Async task doesn't allow a second `configure` call, must use dspy.context(...) instead.
        is_async_task = False
        try:
            if asyncio.current_task() is not None:
                is_async_task = True
        except RuntimeError:
            # This exception (e.g., "no current task") means we are not in an async loop/task,
            # or asyncio module itself is not fully functional in this specific sub-thread context.
            is_async_task = False

        if not is_async_task:
            return

        if config_owner_async_task is None:
            # First `configure` call assigns the owner async task.
            config_owner_async_task = asyncio.current_task()
            return

        # We are in an async task. Now check for IPython and allow calling `configure` from IPython.
        in_ipython = False
        try:
            from IPython import get_ipython

            # get_ipython is a global injected by IPython environments.
            # We check its existence and type to be more robust.
            in_ipython = get_ipython() is not None
        except Exception:
            # If `IPython` is not installed or `get_ipython` failed, we are not in an IPython environment.
            in_ipython = False

        if not in_ipython and config_owner_async_task != asyncio.current_task():
            raise RuntimeError(
                "dspy.configure(...) can only be called from the same async task that called it first. Please "
                "use `dspy.context(...)` in other async tasks instead."
            )

    def configure(self, **kwargs):
        """Set the default language model, adapter, and other settings for DSPy.

        Call `dspy.configure(...)` once near the top of your script. Every
        DSPy module will use these defaults unless you override them with
        `dspy.context(...)`. The values persist until you call
        `dspy.configure(...)` again.

        Only the thread that first calls `dspy.configure(...)` may call it
        again. Use `dspy.context(...)` for temporary overrides in other
        threads or async tasks.

        Args:
            **kwargs: Settings to update. Common keys include `lm` (a
                `dspy.LM`), `adapter` (e.g. `dspy.JSONAdapter()`),
                `callbacks`, `track_usage`, `async_max_workers`, and
                `num_threads`.

        Examples:
            Set a default LM:
            ```python
            import dspy

            dspy.configure(lm=dspy.LM("openai/gpt-5-mini"))
            ```

            Set multiple defaults at once:
            ```python
            import dspy

            dspy.configure(
                lm=dspy.LM("anthropic/claude-sonnet-4-6"),
                adapter=dspy.JSONAdapter(),
                track_usage=True,
            )
            ```

        See Also:
            [`dspy.LM`][dspy.LM]: create the language model you pass as `lm`.
            `dspy.context`: temporary overrides inside one block.
        """
        # `dspy.configure` is documented manually in docs/docs/api/utils/context.md
        # changes here should be reflected there as well.
        # If no exception is raised, the `configure` call is allowed.
        self._ensure_configure_allowed()

        # Update global config
        for k, v in kwargs.items():
            main_thread_config[k] = v

    @contextmanager
    def context(self, **kwargs):
        """Override DSPy settings for one `with` block.

        Use `dspy.context(...)` when you need temporary settings—a different
        LM, adapter, or flag—without changing the process-wide defaults from
        `dspy.configure(...)`. The block inherits every current setting,
        overrides only the keys you pass, and restores the originals on exit.

        Unlike `dspy.configure(...)`, you can call `dspy.context(...)` from
        any thread or async task.

        Args:
            **kwargs: Settings to override, such as `lm`, `adapter`,
                `track_usage`, or `allow_tool_async_sync_conversion`.

        Examples:
            Use a different LM for one call:
            ```python
            import dspy

            dspy.configure(lm=dspy.LM("openai/gpt-5-mini"))
            qa = dspy.Predict("question -> answer")

            with dspy.context(lm=dspy.LM("anthropic/claude-sonnet-4-6")):
                result = qa(question="What is the capital of France?")
                # uses claude-sonnet-4-6 inside this block
            # back to gpt-5-mini here
            ```

        See Also:
            `dspy.configure`: set process-wide defaults.
        """
        # `dspy.context` is documented manually in docs/docs/api/utils/context.md
        # changes here should be reflected there as well.
        original_overrides = thread_local_overrides.get().copy()
        new_overrides = dotdict({**main_thread_config, **original_overrides, **kwargs})
        token = thread_local_overrides.set(new_overrides)

        try:
            yield
        finally:
            thread_local_overrides.reset(token)

    def __repr__(self):
        overrides = thread_local_overrides.get()
        combined_config = {**main_thread_config, **overrides}
        return repr(combined_config)

    def save(
        self, path: str,
        modules_to_serialize: list[str] | None = None,
        exclude_keys: list[str] | None = None,
    ):
        """
        Save the settings to a file using cloudpickle.

        Args:
            path: The file path to save the settings to.
            modules_to_serialize (list or None): A list of modules to serialize with cloudpickle's `register_pickle_by_value`.
                If None, then no modules will be registered for serialization.
            exclude_keys (list or None): A list of keys to exclude during saving.
        """
        logger.warning(
            "`dspy.settings` are serialized using cloudpickle. Because cloudpickle allows for the "
            "execution of arbitrary code during deserialization, you should only load files from "
            "verified sources within a trusted environment."
        )
        try:
            modules_to_serialize = modules_to_serialize or []
            for module in modules_to_serialize:
                cloudpickle.register_pickle_by_value(module)

            exclude_keys = exclude_keys or []
            data = {key: value for key, value in self.config.items() if key not in exclude_keys}
            with open(path, "wb") as f:
                cloudpickle.dump(data, f)
        except Exception as e:
            raise RuntimeError(
                f"Saving failed with error: {e}. Please remove the non-picklable attributes from the values "
                "in the `dspy.settings`."
            )

    @classmethod
    def load(cls, path: str, allow_pickle: bool = False) -> dict[str, Any]:
        """
        Load the settings from a file using cloudpickle.

        Args:
            path: The file path to load the settings from.
            allow_pickle: Whether to allow loading with pickle. Loading untrusted .pkl files
                can run arbitrary code. Set to True only if you trust the source of the file.

        Returns:
            A dict that stores the loaded settings.
        """
        if not allow_pickle:
            raise ValueError(
                "Loading .pkl files can run arbitrary code, which may be dangerous. "
                "Set `allow_pickle=True` if you trust the source of the file."
            )

        with open(path, "rb") as f:
            configs = cloudpickle.load(f)

        return configs


settings = Settings()



from pydantic import BaseModel


class Example:
    """A flexible data container for DSPy examples and training data with named fields.

    An `Example` is roughly one row from a HuggingFace dataset or pandas
    `DataFrame`. It behaves a lot like a dictionary or dot-access record: you
    can read fields with `example["question"]` or `example.question`.

    In DSPy, lists of `Example` objects are your trainset, devset, and testset.
    Most examples are built from keyword arguments or an existing record, then
    tagged with `with_inputs(...)` to say which fields should be fed into a
    module. The remaining fields are labels or metadata.

    When you write evaluation code, custom optimizers, or training loops, use
    `example.inputs()` for the fields you want to pass to a module, and use
    `example.labels()` for the fields you want to compare against the module's
    output.

    Examples:
        Build one from keyword arguments:

        >>> import dspy
        >>> example = dspy.Example(
        ...     question="What is the capital of France?",
        ...     answer="Paris",
        ... ).with_inputs("question")
        >>> example.question
        'What is the capital of France?'
        >>> example.answer
        'Paris'
        >>> example.inputs().toDict()
        {'question': 'What is the capital of France?'}

        Build one from an existing record:

        >>> record = {"question": "What is 2+2?", "answer": "4"}
        >>> example = dspy.Example(**record).with_inputs("question")
        >>> example["question"]
        'What is 2+2?'
        >>> example.labels().answer
        '4'

        Mark which fields are inputs:

        >>> example = dspy.Example(
        ...     question="What is the weather?",
        ...     answer="It's sunny",
        ... ).with_inputs("question")
        >>> example.inputs().question
        'What is the weather?'
        >>> example.labels().answer
        "It's sunny"

        Use examples in a trainset:

        >>> trainset = [
        ...     dspy.Example(question="What is 2+2?", answer="4").with_inputs("question"),
        ...     dspy.Example(question="What is 3+3?", answer="6").with_inputs("question"),
        ... ]
        >>> trainset[0].inputs().toDict()
        {'question': 'What is 2+2?'}

        Use an example in a metric:

        >>> def exact_match_metric(example, pred, trace=None):
        ...     return example.answer.lower() == pred.answer.lower()
        >>> gold = dspy.Example(question="What is 1+1?", answer="2").with_inputs("question")
        >>> pred = dspy.Prediction(answer="2")
        >>> exact_match_metric(gold, pred)
        True

        Use it like a dictionary:

        >>> example = dspy.Example(name="Alice", age=30).with_inputs("name")
        >>> "name" in example
        True
        >>> example.get("city", "Unknown")
        'Unknown'

    See Also:
        [`dspy.Evaluate`][dspy.Evaluate]: Evaluate a program on a list of
            `Example`s.
        [`Metrics`](../../learn/evaluation/metrics.md): Write metric functions
            that compare an `Example` with a prediction.
    """
    # TODO: Add a `dspy.datasets.DataLoader` API link once the datasets API
    # pages exist.

    def __init__(self, base=None, **kwargs):
        """Create an `Example` from fields or from an existing record.

        In the common case, pass fields as keyword arguments, like
        `dspy.Example(question="...", answer="...")`. Use `base` when you
        already have a dictionary or another `Example` and want to copy its
        fields before adding or overriding a few values.

        Args:
            base: A dictionary or `Example` to copy fields from before applying
                `**kwargs`. When `None`, starts with no fields.
            **kwargs: Field names and values to store on the example. If a field
                appears in both `base` and `**kwargs`, the value from
                `**kwargs` wins.
        """
        # Internal storage and other attributes
        self._store = {}
        self._demos = []
        self._input_keys = None

        # Initialize from a base Example if provided
        if base and isinstance(base, type(self)):
            self._store = base._store.copy()

        # Initialize from a dict if provided
        elif base and isinstance(base, dict):
            self._store = base.copy()

        # Update with provided kwargs
        self._store.update(kwargs)

    def __getattr__(self, key):
        if key.startswith("__") and key.endswith("__"):
            raise AttributeError
        if key in self._store:
            return self._store[key]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        if key.startswith("_") or key in dir(self.__class__):
            super().__setattr__(key, value)
        else:
            self._store[key] = value

    def __getitem__(self, key):
        return self._store[key]

    def __setitem__(self, key, value):
        self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __contains__(self, key):
        return key in self._store

    def __len__(self):
        return len([k for k in self._store if not k.startswith("dspy_")])

    def __repr__(self):
        # return f"Example({self._store})" + f" (input_keys={self._input_keys}, demos={self._demos})"
        d = {k: v for k, v in self._store.items() if not k.startswith("dspy_")}
        return f"Example({d})" + f" (input_keys={self._input_keys})"

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Example) and self._store == other._store

    def __hash__(self):
        return hash(tuple(self._store.items()))

    def keys(self, include_dspy=False):
        """Return field names, like `dict.keys()`.

        Args:
            include_dspy: If `True`, include internal fields prefixed
                with `dspy_`. Normally you can ignore these.

        Examples:
            >>> import dspy
            >>> dspy.Example(question="Why?", answer="Because.").keys()
            ['question', 'answer']
        """
        return [k for k in self._store.keys() if not k.startswith("dspy_") or include_dspy]

    def values(self, include_dspy=False):
        """Return field values, like `dict.values()`.

        Args:
            include_dspy: If `True`, include internal fields prefixed
                with `dspy_`.

        Examples:
            >>> import dspy
            >>> dspy.Example(question="Why?", answer="Because.").values()
            ['Why?', 'Because.']
        """
        return [v for k, v in self._store.items() if not k.startswith("dspy_") or include_dspy]

    def items(self, include_dspy=False):
        """Return `(field_name, value)` pairs, like `dict.items()`.

        Args:
            include_dspy: If `True`, include internal fields prefixed
                with `dspy_`.

        Examples:
            >>> import dspy
            >>> dspy.Example(question="Why?", answer="Because.").items()
            [('question', 'Why?'), ('answer', 'Because.')]
        """
        return [(k, v) for k, v in self._store.items() if not k.startswith("dspy_") or include_dspy]

    def get(self, key, default=None):
        """Return the value for `key`, or `default` if the field doesn't exist.

        Args:
            key: Field name to look up.
            default: Value to return when `key` is missing.

        Examples:
            >>> import dspy
            >>> ex = dspy.Example(name="Alice")
            >>> ex.get("name")
            'Alice'
            >>> ex.get("city", "Unknown")
            'Unknown'
        """
        return self._store.get(key, default)

    def with_inputs(self, *keys):
        """Mark which fields are inputs and return a new `Example`.

        Fields not listed here are treated as labels (expected outputs).
        DSPy optimizers and evaluators use this split: they pass
        `example.inputs()` to your program and compare the output against
        `example.labels()`.

        Args:
            *keys: Names of the input fields.

        Returns:
            A copy of this `Example` with the input keys set.

        Examples:
            >>> import dspy
            >>> ex = dspy.Example(question="Why?", answer="Because.").with_inputs("question")
            >>> ex.inputs().keys()
            ['question']
            >>> ex.labels().keys()
            ['answer']
        """
        copied = self.copy()
        copied._input_keys = set(keys)
        return copied

    def inputs(self):
        """Return a new `Example` containing only the input fields.

        Requires `with_inputs` to have been called first.

        Raises:
            ValueError: If `with_inputs` was not called on this example.

        Examples:
            >>> import dspy
            >>> ex = dspy.Example(question="Why?", answer="Because.").with_inputs("question")
            >>> ex.inputs()
            Example({'question': 'Why?'}) (input_keys={'question'})
        """
        if self._input_keys is None:
            raise ValueError("Inputs have not been set for this example. Use `example.with_inputs()` to set them.")

        # return items that are in input_keys
        d = {key: self._store[key] for key in self._store if key in self._input_keys}
        # return type(self)(d)
        new_instance = type(self)(base=d)
        new_instance._input_keys = self._input_keys  # Preserve input_keys in new instance
        return new_instance

    def labels(self):
        """Return a new `Example` containing only the label (non-input) fields.

        Requires `with_inputs` to have been called first, since labels are
        everything that is *not* an input.

        Examples:
            >>> import dspy
            >>> ex = dspy.Example(question="Why?", answer="Because.").with_inputs("question")
            >>> ex.labels()
            Example({'answer': 'Because.'}) (input_keys=None)
        """
        # return items that are NOT in input_keys
        input_keys = self.inputs().keys()
        d = {key: self._store[key] for key in self._store if key not in input_keys}
        return type(self)(d)

    def __iter__(self):
        return iter(dict(self._store))

    def copy(self, **kwargs):
        """Return a shallow copy, optionally overriding fields.

        Args:
            **kwargs: Fields to add or override in the copy.

        Examples:
            >>> import dspy
            >>> ex = dspy.Example(question="Why?", answer="Because.")
            >>> ex.copy(answer="No reason.")
            Example({'question': 'Why?', 'answer': 'No reason.'}) (input_keys=None)
        """
        return type(self)(base=self, **kwargs)

    def without(self, *keys):
        """Return a copy with the specified fields removed.

        Args:
            *keys: Field names to drop.

        Examples:
            >>> import dspy
            >>> ex = dspy.Example(question="Why?", answer="Because.", source="web")
            >>> ex.without("source")
            Example({'question': 'Why?', 'answer': 'Because.'}) (input_keys=None)
        """
        copied = self.copy()
        for key in keys:
            del copied[key]
        return copied

    def toDict(self):  # noqa: N802
        """Convert to a plain dictionary, recursively serializing nested objects.

        Nested `Example` objects, Pydantic models, lists, and dicts are
        converted so the result is JSON-friendly.

        Examples:
            >>> import dspy
            >>> dspy.Example(question="Why?", answer="Because.").toDict()
            {'question': 'Why?', 'answer': 'Because.'}
        """
        def convert_to_serializable(value):
            if hasattr(value, "toDict"):
                return value.toDict()
            elif isinstance(value, BaseModel):
                # Handle Pydantic models (e.g., dspy.History)
                return value.model_dump()
            elif isinstance(value, list):
                return [convert_to_serializable(item) for item in value]
            elif isinstance(value, dict):
                return {k: convert_to_serializable(v) for k, v in value.items()}
            else:
                return value

        serializable_store = {}
        for k, v in self._store.items():
            serializable_store[k] = convert_to_serializable(v)

        return serializable_store



import logging
import random
from typing import Any, Callable, Protocol, TypedDict

from gepa import EvaluationBatch, GEPAAdapter
from gepa.core.adapter import ProposalFn
from gepa.strategies.instruction_proposal import InstructionProposalSignature

import dspy
from dspy.adapters.chat_adapter import ChatAdapter
from dspy.adapters.types import History
from dspy.adapters.types.base_type import Type
from dspy.evaluate import Evaluate
from dspy.primitives import Example, Prediction
from dspy.teleprompt.bootstrap_trace import FailedPrediction, TraceData

logger = logging.getLogger(__name__)


class LoggerAdapter:
    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def log(self, x: str):
        self.logger.info(x)


DSPyTrace = list[tuple[Any, dict[str, Any], Prediction]]

ReflectiveExample = TypedDict(
    "ReflectiveExample",
    {
        "Inputs": dict[str, Any],
        "Generated Outputs": dict[str, Any] | str,
        "Feedback": str,
    },
)

ReflectiveExample.__doc__ = """
Structure of individual examples in the reflective dataset.

Each example contains the predictor inputs, generated outputs, and feedback from evaluation.
"""


class ScoreWithFeedback(Prediction):
    score: float
    feedback: str


class PredictorFeedbackFn(Protocol):
    def __call__(
        self,
        predictor_output: dict[str, Any],
        predictor_inputs: dict[str, Any],
        module_inputs: Example,
        module_outputs: Prediction,
        captured_trace: DSPyTrace,
    ) -> ScoreWithFeedback:
        """
        This function is used to provide feedback to a specific predictor.
        The function is called with the following arguments:
        - predictor_output: The output of the predictor.
        - predictor_inputs: The inputs to the predictor.
        - module_inputs: The inputs to the whole program --- `Example`.
        - module_outputs: The outputs of the whole program --- `Prediction`.
        - captured_trace: The trace of the module's execution.
        # Shape of trace is: [predictor_invocation_idx -> Tuple[Predictor, PredictorInputs, Prediction]]
        # Each trace is a tuple of (Predictor, PredictorInputs, Prediction)

        The function should return a `ScoreWithFeedback` object.
        The feedback is a string that is used to guide the evolution of the predictor.
        """
        ...


class DspyAdapter(GEPAAdapter[Example, TraceData, Prediction]):
    def __init__(
        self,
        student_module,
        metric_fn: Callable,
        feedback_map: dict[str, Callable],
        failure_score=0.0,
        num_threads: int | None = None,
        add_format_failure_as_feedback: bool = False,
        rng: random.Random | None = None,
        reflection_lm=None,
        custom_instruction_proposer: "ProposalFn | None" = None,
        warn_on_score_mismatch: bool = True,
        reflection_minibatch_size: int | None = None,
    ):
        self.student = student_module
        self.metric_fn = metric_fn
        self.feedback_map = feedback_map
        self.failure_score = failure_score
        self.num_threads = num_threads
        self.add_format_failure_as_feedback = add_format_failure_as_feedback
        self.rng = rng or random.Random(0)
        self.reflection_lm = reflection_lm
        self.custom_instruction_proposer = custom_instruction_proposer
        self.warn_on_score_mismatch = warn_on_score_mismatch
        self.reflection_minibatch_size = reflection_minibatch_size

    def propose_new_texts(
        self,
        candidate: dict[str, str],
        reflective_dataset: dict[str, list[dict[str, Any]]],
        components_to_update: list[str],
    ) -> dict[str, str]:
        reflection_lm = self.reflection_lm or dspy.settings.lm
        # If custom proposer provided, override everything with custom proposer
        if self.custom_instruction_proposer:
            with dspy.context(lm=reflection_lm):
                return self.custom_instruction_proposer(
                    candidate=candidate,
                    reflective_dataset=reflective_dataset,
                    components_to_update=components_to_update,
                )

        results: dict[str, str] = {}

        with dspy.context(lm=reflection_lm):
            for name in components_to_update:
                base_instruction = candidate[name]
                dataset_with_feedback = reflective_dataset[name]
                results[name] = InstructionProposalSignature.run(
                    lm=(lambda x: self.stripped_lm_call(x)[0]),
                    input_dict={
                        "current_instruction_doc": base_instruction,
                        "dataset_with_feedback": dataset_with_feedback,
                    },
                )["new_instruction"]

        return results

    def build_program(self, candidate: dict[str, str]):
        new_prog = self.student.deepcopy()

        for name, pred in new_prog.named_predictors():
            if name in candidate:
                pred.signature = pred.signature.with_instructions(candidate[name])

        return new_prog

    def evaluate(self, batch, candidate, capture_traces=False):
        program = self.build_program(candidate)
        callback_metadata = (
            {"metric_key": "eval_full"}
            if self.reflection_minibatch_size is None or len(batch) > self.reflection_minibatch_size
            else {"disable_logging": True}
        )

        if capture_traces:
            # bootstrap_trace_data-like flow with trace capture
            from dspy.teleprompt import bootstrap_trace as bootstrap_trace_module

            trajs = bootstrap_trace_module.bootstrap_trace_data(
                program=program,
                dataset=batch,
                metric=self.metric_fn,
                num_threads=self.num_threads,
                raise_on_error=False,
                capture_failed_parses=True,
                failure_score=self.failure_score,
                format_failure_score=self.failure_score,
                callback_metadata=callback_metadata,
            )
            scores = []
            outputs = []
            for t in trajs:
                outputs.append(t["prediction"])
                if hasattr(t["prediction"], "__class__") and t.get("score") is None:
                    scores.append(self.failure_score)
                else:
                    score = t["score"]
                    if hasattr(score, "score"):
                        score = score["score"]
                    scores.append(score)

            return EvaluationBatch(outputs=outputs, scores=scores, trajectories=trajs)
        else:
            evaluator = Evaluate(
                devset=batch,
                metric=self.metric_fn,
                num_threads=self.num_threads,
                return_all_scores=True,
                failure_score=self.failure_score,
                provide_traceback=True,
                max_errors=len(batch) * 100,
                callback_metadata=callback_metadata,
            )
            res = evaluator(program)
            outputs = [r[1] for r in res.results]
            scores = [r[2] for r in res.results]
            scores = [s["score"] if hasattr(s, "score") else s for s in scores]
            return EvaluationBatch(outputs=outputs, scores=scores, trajectories=None)

    def make_reflective_dataset(
        self, candidate, eval_batch, components_to_update
    ) -> dict[str, list[ReflectiveExample]]:
        program = self.build_program(candidate)

        ret_d: dict[str, list[ReflectiveExample]] = {}

        for pred_name in components_to_update:
            # Find the predictor object
            module = None
            for name, m in program.named_predictors():
                if name == pred_name:
                    module = m
                    break
            assert module is not None, f"Predictor not found: {pred_name}"

            # Create reflective examples from traces
            items: list[ReflectiveExample] = []
            for data in eval_batch.trajectories or []:
                trace = data["trace"]
                example = data["example"]
                prediction = data["prediction"]
                module_score = data["score"]
                if hasattr(module_score, "score"):
                    module_score = module_score["score"]

                trace_instances = [t for t in trace if t[0].signature.equals(module.signature)]
                if not self.add_format_failure_as_feedback:
                    trace_instances = [t for t in trace_instances if not isinstance(t[2], FailedPrediction)]
                if len(trace_instances) == 0:
                    continue

                selected = None
                for t in trace_instances:
                    if isinstance(t[2], FailedPrediction):
                        selected = t
                        break

                if selected is None:
                    if isinstance(prediction, FailedPrediction):
                        continue
                    selected = self.rng.choice(trace_instances)

                inputs = selected[1]
                outputs = selected[2]

                new_inputs = {}
                new_outputs = {}

                contains_history = False
                history_key_name = None
                for input_key, input_val in inputs.items():
                    if isinstance(input_val, History):
                        contains_history = True
                        assert history_key_name is None
                        history_key_name = input_key

                if contains_history:
                    s = "```json\n"
                    for i, message in enumerate(inputs[history_key_name].messages):
                        s += f"  {i}: {message}\n"
                    s += "```"
                    new_inputs["Context"] = s

                for input_key, input_val in inputs.items():
                    if contains_history and input_key == history_key_name:
                        continue

                    if isinstance(input_val, Type) and self.custom_instruction_proposer is not None:
                        # Keep original object - will be properly formatted when sent to reflection LM
                        new_inputs[input_key] = input_val
                    else:
                        new_inputs[input_key] = str(input_val)

                if isinstance(outputs, FailedPrediction):
                    s = "Couldn't parse the output as per the expected output format. The model's raw response was:\n"
                    s += "```\n"
                    s += outputs.completion_text + "\n"
                    s += "```\n\n"
                    new_outputs = s
                else:
                    for output_key, output_val in outputs.items():
                        new_outputs[output_key] = str(output_val)

                d = {"Inputs": new_inputs, "Generated Outputs": new_outputs}
                if isinstance(outputs, FailedPrediction):
                    adapter = ChatAdapter()
                    structure_instruction = ""
                    for dd in adapter.format(module.signature, [], {}):
                        structure_instruction += dd["role"] + ": " + dd["content"] + "\n"
                    d["Feedback"] = "Your output failed to parse. Follow this structure:\n" + structure_instruction
                    # d['score'] = self.failure_score
                else:
                    feedback_fn = self.feedback_map[pred_name]
                    fb = feedback_fn(
                        predictor_output=outputs,
                        predictor_inputs=inputs,
                        module_inputs=example,
                        module_outputs=prediction,
                        captured_trace=trace,
                    )
                    d["Feedback"] = fb["feedback"]
                    if fb["score"] != module_score:
                        if self.warn_on_score_mismatch:
                            logger.warning(
                                "The score returned by the metric with pred_name is different from the overall metric score. This can indicate 2 things: Either the metric is non-deterministic (e.g., LLM-as-judge, Semantic score, etc.) or the metric returned a score specific to pred_name that differs from the module level score. Currently, GEPA does not support predictor level scoring (support coming soon), and only requires a feedback text to be provided, which can be specific to the predictor or program level. GEPA will ignore the differing score returned, and instead use module level score. You can safely ignore this warning if using a semantic metric, however, if this mismatch is caused due to predictor scoring, please return module-level scores. To disable this warning, set warn_on_score_mismatch=False."
                            )
                            self.warn_on_score_mismatch = False
                        fb["score"] = module_score

                items.append(d)

            if len(items) == 0:
                logger.warning(f"  No valid reflective examples found for {pred_name}")
                continue

            ret_d[pred_name] = items

        if len(ret_d) == 0:
            raise Exception("No valid predictions found for any module.")

        return ret_d

    # Always return strings from the LM outputs
    # Even when it returns a dict with e.g., "text" and "reasoning" fields
    def stripped_lm_call(self, x: str) -> list[str]:
        raw_outputs = self.reflection_lm(x)
        outputs = []
        for raw_output in raw_outputs:
            if type(raw_output) == str:
                outputs.append(raw_output)
            elif type(raw_output) == dict:
                if "text" not in raw_output:
                    raise KeyError("Missing 'text' field in the output from the base LM!")
                outputs.append(raw_output["text"])
            else:
                raise TypeError("Unexpected output type from the base LM! Expected str or dict")

        return outputs




#replace_package_name_marker
__name__="dspy"
__package_name__="dspy-lite"
#replace_package_version_marker
__version__="3.1.3.post1"
__description__="DSPy with litellm replaced by litelm and numpy made optional"
__url__="https://github.com/kennethwolters/dspy-lite"
__author__="Kenneth Wolters"
__author_email__=""



import logging
import random
from typing import Any, Literal, get_args, get_origin

from pydantic import BaseModel
from pydantic_core import PydanticUndefined
from typeguard import TypeCheckError, check_type

from dspy.adapters.chat_adapter import ChatAdapter
from dspy.clients.base_lm import BaseLM
from dspy.clients.lm import LM
from dspy.dsp.utils.settings import settings
from dspy.predict.parameter import Parameter
from dspy.primitives.module import Module
from dspy.primitives.prediction import Prediction
from dspy.signatures.signature import Signature, ensure_signature
from dspy.utils.callback import BaseCallback
from dspy.utils.constants import IS_TYPE_UNDEFINED

logger = logging.getLogger(__name__)

UNSAFE_LM_STATE_KEYS = {"api_base", "base_url"}


def _sanitize_lm_state(lm_state: dict, allow_unsafe_lm_state: bool) -> dict:
    if allow_unsafe_lm_state:
        return lm_state

    unsafe_keys = sorted(UNSAFE_LM_STATE_KEYS.intersection(lm_state))

    if not unsafe_keys:
        return lm_state

    sanitized_lm_state = {k: v for k, v in lm_state.items() if k not in UNSAFE_LM_STATE_KEYS}
    logger.warning(
        "Ignoring unsafe LM config key(s) during state load: %s. "
        "Pass allow_unsafe_lm_state=True to preserve these keys for trusted files.",
        unsafe_keys,
    )
    return sanitized_lm_state


class Predict(Module, Parameter):
    """Basic DSPy module that maps inputs to outputs using a language model.

    Args:
        signature: The input/output signature describing the task.
        callbacks: Optional list of callbacks for instrumentation.
        **config: Default keyword arguments forwarded to the underlying
            language model. These values can be overridden for a single
            invocation by passing a ``config`` dictionary when calling the
            module. For example::

                predict = dspy.Predict("q -> a", rollout_id=1, temperature=1.0)
                predict(q="What is 1 + 52?", config={"rollout_id": 2, "temperature": 1.0})
    """

    def __init__(self, signature: str | type[Signature], callbacks: list[BaseCallback] | None = None, **config):
        super().__init__(callbacks=callbacks)
        self.stage = random.randbytes(8).hex()
        self.signature = ensure_signature(signature)
        self.config = config
        self.reset()

    def reset(self):
        self.lm = None
        self.traces = []
        self.train = []
        self.demos = []

    def dump_state(self, json_mode=True):
        state_keys = ["traces", "train"]
        state = {k: getattr(self, k) for k in state_keys}

        state["demos"] = []
        for demo in self.demos:
            demo = demo.copy()

            for field in demo:
                # FIXME: Saving BaseModels as strings in examples doesn't matter because you never re-access as an object
                demo[field] = serialize_object(demo[field])

            if isinstance(demo, dict) or not json_mode:
                state["demos"].append(demo)
            else:
                state["demos"].append(demo.toDict())

        state["signature"] = self.signature.dump_state()
        state["lm"] = self.lm.dump_state() if self.lm else None
        return state

    def load_state(self, state: dict, *, allow_unsafe_lm_state: bool = False) -> "Predict":
        """Load the saved state of a `Predict` object.

        Args:
            state: The saved state of a `Predict` object.
            allow_unsafe_lm_state: If True, preserves `api_base` and `base_url` from
                serialized LM state. Enable only when loading trusted files.

        Returns:
            Self to allow method chaining.
        """
        excluded_keys = ["signature", "extended_signature", "lm"]
        for name, value in state.items():
            # `excluded_keys` are fields that go through special handling.
            if name not in excluded_keys:
                setattr(self, name, value)

        self.signature = self.signature.load_state(state["signature"])
        sanitized_lm_state = _sanitize_lm_state(state["lm"], allow_unsafe_lm_state) if state["lm"] else None
        self.lm = LM(**sanitized_lm_state) if sanitized_lm_state else None

        if "extended_signature" in state:  # legacy, up to and including 2.5, for CoT.
            raise NotImplementedError("Loading extended_signature is no longer supported in DSPy 2.6+")

        return self

    def _get_positional_args_error_message(self):
        input_fields = list(self.signature.input_fields.keys())
        return (
            "Positional arguments are not allowed when calling `dspy.Predict`, must use keyword arguments "
            f"that match your signature input fields: '{', '.join(input_fields)}'. For example: "
            f"`predict({input_fields[0]}=input_value, ...)`."
        )

    def __call__(self, *args, **kwargs):
        if args:
            raise ValueError(self._get_positional_args_error_message())

        return super().__call__(**kwargs)

    async def acall(self, *args, **kwargs):
        if args:
            raise ValueError(self._get_positional_args_error_message())

        return await super().acall(**kwargs)

    def _forward_preprocess(self, **kwargs):
        # Extract the three privileged keyword arguments.
        assert "new_signature" not in kwargs, "new_signature is no longer a valid keyword argument."
        signature = ensure_signature(kwargs.pop("signature", self.signature))
        demos = kwargs.pop("demos", self.demos)
        config = {**self.config, **kwargs.pop("config", {})}

        # Get the right LM to use.
        lm = kwargs.pop("lm", self.lm) or settings.lm

        if lm is None:
            raise ValueError(
                "No LM is loaded. Please configure the LM using `dspy.configure(lm=dspy.LM(...))`. e.g, "
                "`dspy.configure(lm=dspy.LM('openai/gpt-4o-mini'))`"
            )

        if isinstance(lm, str):
            # Many users mistakenly use `dspy.configure(lm="openai/gpt-4o-mini")` instead of
            # `dspy.configure(lm=dspy.LM("openai/gpt-4o-mini"))`, so we are providing a specific error message.
            raise ValueError(
                f"LM must be an instance of `dspy.BaseLM`, not a string. Instead of using a string like "
                f"'dspy.configure(lm=\"{lm}\")', please configure the LM like 'dspy.configure(lm=dspy.LM(\"{lm}\"))'"
            )
        elif not isinstance(lm, BaseLM):
            raise ValueError(f"LM must be an instance of `dspy.BaseLM`, not {type(lm)}. Received `lm={lm}`.")

        # If temperature is unset or <=0.15, and n > 1, set temperature to 0.7 to keep randomness.
        temperature = config.get("temperature") or lm.kwargs.get("temperature")
        num_generations = config.get("n") or lm.kwargs.get("n") or lm.kwargs.get("num_generations") or 1

        if (temperature is None or temperature <= 0.15) and num_generations > 1:
            config["temperature"] = 0.7

        if "prediction" in kwargs:
            if (
                isinstance(kwargs["prediction"], dict)
                and kwargs["prediction"].get("type") == "content"
                and "content" in kwargs["prediction"]
            ):
                # If the `prediction` is the standard predicted outputs format
                # (https://platform.openai.com/docs/guides/predicted-outputs), we remove it from input kwargs and add it
                # to the lm kwargs.
                config["prediction"] = kwargs.pop("prediction")

        # Populate default values for missing input fields.
        for k, v in signature.input_fields.items():
            if k not in kwargs and v.default is not PydanticUndefined:
                kwargs[k] = v.default

        # Check and warn for extra fields not in signature
        extra_fields = [k for k in kwargs if k not in signature.input_fields]
        if extra_fields:
            logger.warning(
                "Input contains fields not in signature. These fields will be ignored: %s. "
                "Expected fields: %s.",
                extra_fields,
                list(signature.input_fields.keys()),
            )

        # Validate input field types match signature
        if settings.warn_on_type_mismatch:
            for field_name, field_info in signature.input_fields.items():
                if field_name in kwargs:
                    value = kwargs[field_name]
                    expected_type: type = field_info.annotation

                    if value is None or field_info.json_schema_extra.get(IS_TYPE_UNDEFINED, False):
                        continue

                    if not _is_value_compatible_with_type(value, expected_type):
                        logger.warning(
                            "Type mismatch for field '%s': expected %s based on given Signature, "
                            "but the provided value is incompatible: %s.",
                            field_name,
                            _get_type_name(expected_type),
                            value,
                        )

        if not all(k in kwargs for k in signature.input_fields):
            present = [k for k in signature.input_fields if k in kwargs]
            missing = [k for k in signature.input_fields if k not in kwargs]
            logger.warning(
                "Not all input fields were provided to module. Present: %s. Missing: %s.",
                present,
                missing,
            )
        return lm, config, signature, demos, kwargs

    def _forward_postprocess(self, completions, signature, **kwargs):
        pred = Prediction.from_completions(completions, signature=signature)
        if kwargs.pop("_trace", True) and settings.trace is not None and settings.max_trace_size > 0:
            trace = settings.trace
            if len(trace) >= settings.max_trace_size:
                trace.pop(0)
            trace.append((self, {**kwargs}, pred))
        return pred

    def _should_stream(self):
        stream_listeners = settings.stream_listeners or []
        should_stream = settings.send_stream is not None
        if should_stream and len(stream_listeners) > 0:
            should_stream = any(stream_listener.predict == self for stream_listener in stream_listeners)

        return should_stream

    def forward(self, **kwargs):
        lm, config, signature, demos, kwargs = self._forward_preprocess(**kwargs)

        adapter = settings.adapter or ChatAdapter()

        if self._should_stream():
            with settings.context(caller_predict=self):
                completions = adapter(lm, lm_kwargs=config, signature=signature, demos=demos, inputs=kwargs)
        else:
            with settings.context(send_stream=None):
                completions = adapter(lm, lm_kwargs=config, signature=signature, demos=demos, inputs=kwargs)

        return self._forward_postprocess(completions, signature, **kwargs)

    async def aforward(self, **kwargs):
        lm, config, signature, demos, kwargs = self._forward_preprocess(**kwargs)

        adapter = settings.adapter or ChatAdapter()
        if self._should_stream():
            with settings.context(caller_predict=self):
                completions = await adapter.acall(lm, lm_kwargs=config, signature=signature, demos=demos, inputs=kwargs)
        else:
            with settings.context(send_stream=None):
                completions = await adapter.acall(lm, lm_kwargs=config, signature=signature, demos=demos, inputs=kwargs)

        return self._forward_postprocess(completions, signature, **kwargs)

    def update_config(self, **kwargs):
        self.config = {**self.config, **kwargs}

    def get_config(self):
        return self.config

    def __repr__(self):
        return f"{self.__class__.__name__}({self.signature})"

def _get_type_name(type_annotation) -> str:
    """Helper method to get the name for a type annotation."""

    origin = get_origin(type_annotation)
    args = get_args(type_annotation)

    if origin is None:
        # Primitives like str, int, etc.
        if hasattr(type_annotation, "__name__"):
            return type_annotation.__name__
        return str(type_annotation)

    # Handle Literal types
    if origin is Literal:
        literal_values = ", ".join(repr(arg) for arg in args)
        return f"Literal[{literal_values}]"

    # Types like list[str], dict[str, int], generics, etc.
    if args:
        # Handle Ellipsis in tuples (e.g., tuple[int, ...])
        args_str = ", ".join("..." if arg is ... else _get_type_name(arg) for arg in args)
        origin_name = getattr(origin, "__name__", str(origin))
        return f"{origin_name}[{args_str}]"

    return getattr(origin, "__name__", str(origin))

def _is_value_compatible_with_type(value: Any, expected: type) -> bool:
    """Return True if the value matches the expected type hint."""
    try:
        # Special handle list[str] because we allow setting input type to str, however, invoking with a list thereof.
        if expected is str and isinstance(value, list):
            if all(isinstance(item, str) for item in value):
                return True

        check_type(value, expected)
        return True
    except TypeCheckError:
        return False

def serialize_object(obj):
    """
    Recursively serialize a given object into a JSON-compatible format.
    Supports Pydantic models, lists, dicts, and primitive types.
    """
    if isinstance(obj, BaseModel):
        # Use model_dump with mode="json" to ensure all fields (including HttpUrl, datetime, etc.)
        # are converted to JSON-serializable types (strings)
        return obj.model_dump(mode="json")
    elif isinstance(obj, list):
        return [serialize_object(item) for item in obj]
    elif isinstance(obj, tuple):
        return tuple(serialize_object(item) for item in obj)
    elif isinstance(obj, dict):
        return {key: serialize_object(value) for key, value in obj.items()}
    else:
        return obj


# # TODO: FIXME: Hmm, I guess expected behavior is that contexts can
# affect execution. Well, we need to determine whether context dominates, __init__ demoninates, or forward dominates.
# Generally, unless overwritten, we'd see n=None, temperature=None.
# That will eventually mean we have to learn them.



from typing import Any, Literal

import pydantic

from dspy.adapters.types.base_type import Type
from dspy.utils.annotation import experimental


@experimental(version="3.0.4")
class Document(Type):
    """A document type for providing content that can be cited by language models.

    This type represents documents that can be passed to language models for citation-enabled
    responses, particularly useful with Anthropic's Citations API. Documents include the content
    and metadata that helps the LM understand and reference the source material.

    Attributes:
        data: The text content of the document
        title: Optional title for the document (used in citations)
        media_type: MIME type of the document content (defaults to "text/plain")
        context: Optional context information about the document

    Examples:
        ```python
        import dspy
        from dspy.signatures import Signature
        from dspy.experimental import Document, Citations

        class AnswerWithSources(Signature):
            '''Answer questions using provided documents with citations.'''
            documents: list[Document] = dspy.InputField()
            question: str = dspy.InputField()
            answer: str = dspy.OutputField()
            citations: Citations = dspy.OutputField()

        # Create documents
        docs = [
            Document(
                data="The Earth orbits the Sun in an elliptical path.",
                title="Basic Astronomy Facts"
            ),
            Document(
                data="Water boils at 100°C at standard atmospheric pressure.",
                title="Physics Fundamentals",
            )
        ]

        # Use with a citation-supporting model
        lm = dspy.LM("anthropic/claude-opus-4-1-20250805")
        predictor = dspy.Predict(AnswerWithSources)
        result = predictor(documents=docs, question="What temperature does water boil?", lm=lm)
        print(result.citations)
        ```
    """

    data: str
    title: str | None = None
    media_type: Literal["text/plain", "application/pdf"] = "text/plain"
    context: str | None = None

    def format(self) -> list[dict[str, Any]]:
        """Format document for LM consumption.

        Returns:
            A list containing the document block in the format expected by citation-enabled language models.
        """
        document_block = {
            "type": "document",
            "source": {
                "type": "text",
                "media_type": self.media_type,
                "data": self.data
            },
            "citations": {"enabled": True}
        }

        if self.title:
            document_block["title"] = self.title

        if self.context:
            document_block["context"] = self.context

        return [document_block]



    @classmethod
    def description(cls) -> str:
        """Description of the document type for use in prompts."""
        return (
            "A document containing text content that can be referenced and cited. "
            "Include the full text content and optionally a title for proper referencing."
        )

    @pydantic.model_validator(mode="before")
    @classmethod
    def validate_input(cls, data: Any):
        if isinstance(data, cls):
            return data

        # Handle case where data is just a string (data only)
        if isinstance(data, str):
            return {"data": data}

        # Handle case where data is a dict
        elif isinstance(data, dict):
            return data

        raise ValueError(f"Received invalid value for `Document`: {data}")

    def __str__(self) -> str:
        """String representation showing title and content length."""
        title_part = f"'{self.title}': " if self.title else ""
        return f"Document({title_part}{len(self.data)} chars)"



# TODO: This should move internally. Same for passage_match. dspy.metrics.answer_exact_match, dspy.metrics.answer_passage_match

import re
import string
import unicodedata
from collections import Counter

from dspy.dsp.utils.utils import print_message


def EM(prediction, answers_list):  # noqa: N802
    """Compute the Exact Match (EM) metric between a prediction and reference answers.

    Returns True if any reference exactly matches the prediction after normalization;
    otherwise False. Normalization applies Unicode NFD, lowercasing, punctuation
    removal, English article removal ("a", "an", "the"), and whitespace collapse.

    Args:
        prediction (str): Predicted answer string.
        answers_list (list[str]): List of reference answers.

    Returns:
        bool: Whether any reference exactly equals the prediction after normalization.

    Examples:
        ```python
        EM("The Eiffel Tower", ["Eiffel Tower", "Louvre"])  # True

        EM("paris", ["Paris"])  # True
        EM("paris", ["Paris, France"])  # False
        ```
    """
    if not isinstance(answers_list, list):
        raise ValueError(f"`answers_list` must be a list, got {type(answers_list)}")

    return max(em_score(prediction, ans) for ans in answers_list)


def F1(prediction, answers_list):  # noqa: N802
    """Compute the maximum token-level F1 score against reference answers.

    Strings are normalized (same as in `EM`) and whitespace-tokenized. The function
    returns the maximum F1 over all provided references.

    Args:
        prediction (str): Predicted answer string.
        answers_list (list[str]): List of reference answers.

    Returns:
        float: Highest F1 score in [0.0, 1.0].

    Examples:
        ```python
        round(F1("Eiffel Tower is in Paris", ["Paris"]), 2)  # 0.33
        ```
    """
    if not isinstance(answers_list, list):
        raise ValueError(f"`answers_list` must be a list, got {type(answers_list)}")

    return max(f1_score(prediction, ans) for ans in answers_list)


def HotPotF1(prediction, answers_list):  # noqa: N802
    """Compute the maximum HotPotQA-style F1 score against reference answers.

    Like `F1`, but if either normalized side is one of {"yes", "no", "noanswer"}
    and they differ, the score is 0. Otherwise, standard token-level F1 is used.

    Args:
        prediction (str): Predicted answer.
        answers_list (list[str]): List of reference answers.

    Returns:
        float: Highest HotPotQA-style F1 in [0.0, 1.0].

    Examples:
        ```python
        HotPotF1("yes", ["no"])  # 0.0
        ```
    """
    if not isinstance(answers_list, list):
        raise ValueError(f"`answers_list` must be a list, got {type(answers_list)}")

    return max(hotpot_f1_score(prediction, ans) for ans in answers_list)


def normalize_text(s):
    """Normalize text for string and token comparisons.

    Steps:
        1) Unicode NFD normalization
        2) lowercasing
        3) punctuation removal
        4) English article removal ("a", "an", "the")
        5) whitespace collapse

    Args:
        s (str): Input string.

    Returns:
        str: Normalized string.

    Examples:
        ```python
        normalize_text("The,  Eiffel  Tower!")  # "eiffel tower"
        ```
    """
    s = unicodedata.normalize("NFD", s)

    def remove_articles(text):
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def em_score(prediction, ground_truth):
    """Compute boolean exact match after normalization.

    Args:
        prediction (str): Predicted answer.
        ground_truth (str): Reference answer.

    Returns:
        bool: True if normalized strings are identical; otherwise False.

    Examples:
        ```python
        em_score("Paris", "paris")  # True
        ```
    """
    return normalize_text(prediction) == normalize_text(ground_truth)


def f1_score(prediction, ground_truth):
    """Compute token-level F1 between prediction and reference (after normalization).

    Strings are normalized (see `normalize_text`) and split by whitespace. F1 is
    computed from token precision and recall. If there is no token overlap, returns 0.
    If both sides are empty, a diagnostic message is printed; score remains 0.

    Args:
        prediction (str): Predicted answer.
        ground_truth (str): Reference answer.

    Returns:
        float: F1 score in [0.0, 1.0].

    Examples:
        ```python
        round(f1_score("the Eiffel Tower", "Eiffel Tower"), 2)  # 1.0
        ```
    """
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print_message("\n#> F1 Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n")

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)

    return f1


def hotpot_f1_score(prediction, ground_truth):
    """Compute HotPotQA-style token F1 with special labels.

    If either normalized string is in {"yes", "no", "noanswer"} and they differ,
    the score is 0. Otherwise compute standard token F1 after normalization.

    Args:
        prediction (str): Predicted answer.
        ground_truth (str): Reference answer.

    Returns:
        float: HotPotQA-style F1 score in [0.0, 1.0].

    Examples:
        ```python
        hotpot_f1_score("no", "yes")  # 0.0
        ```
    """
    normalized_prediction = normalize_text(prediction)
    normalized_ground_truth = normalize_text(ground_truth)

    if normalized_prediction in ["yes", "no", "noanswer"] and normalized_prediction != normalized_ground_truth:
        return 0
    if normalized_ground_truth in ["yes", "no", "noanswer"] and normalized_prediction != normalized_ground_truth:
        return 0

    prediction_tokens = normalized_prediction.split()
    ground_truth_tokens = normalized_ground_truth.split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def precision_score(prediction, ground_truth):
    """Compute token-level precision of prediction against reference (after normalization).

    Precision is (# overlapping tokens) / (# tokens in prediction). If there is no
    token overlap, returns 0. If both sides are empty, a diagnostic message is printed;
    precision remains 0.

    Args:
        prediction (str): Predicted answer.
        ground_truth (str): Reference answer.

    Returns:
        float: Precision in [0.0, 1.0].

    Examples:
        ```python
        precision_score("eiffel tower in paris", "eiffel tower")  # 0.67
        ```
    """
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print_message(
            "\n#> Precision Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n"
        )

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)
    return precision


def _passage_match(passages: list[str], answers: list[str]) -> bool:
    """Return True if any passage contains any answer (normalized & DPR-normalized)."""
    from dspy.dsp.utils import DPR_normalize, has_answer

    def passage_has_answers(passage: str, answers: list[str]) -> bool:
        """Return True if the passage contains any of the answers."""
        return has_answer(
            tokenized_answers=[DPR_normalize(normalize_text(ans)) for ans in answers],
            text=normalize_text(passage),
        )

    return any(passage_has_answers(psg, answers) for psg in passages)


def _answer_match(prediction, answers, frac=1.0):
    """Return True if prediction matches any answer.

    When `frac >= 1.0`, require exact match (EM). Otherwise, return whether the
    maximum token-level F1 across answers is at least `frac`.
    """
    if frac >= 1.0:
        return EM(prediction, answers)

    return F1(prediction, answers) >= frac


def answer_exact_match(example, pred, trace=None, frac=1.0):
    """Evaluate exact match or F1-thresholded match for an example/prediction pair.

    If `example.answer` is a string, compare `pred.answer` against it. If it's a list,
    compare against any of the references. When `frac >= 1.0` (default), use EM;
    otherwise require that the maximum F1 across references is at least `frac`.

    Args:
        example: `dspy.Example` object with field `answer` (str or list[str]).
        pred: `dspy.Prediction` object with field `answer` (str).
        trace: Unused; reserved for compatibility.
        frac (float, optional): Threshold in [0.0, 1.0]. `1.0` means EM.

    Returns:
        bool: True if the match condition holds; otherwise False.

    Examples:
        ```python
        import dspy

        example = dspy.Example(answer=["Eiffel Tower", "Louvre"])
        pred = dspy.Prediction(answer="The Eiffel Tower")

        answer_exact_match(example, pred, frac=1.0)  # equivalent to EM, True
        answer_exact_match(example, pred, frac=0.5)  # True
        ```
    """
    if isinstance(example.answer, str):
        return _answer_match(pred.answer, [example.answer], frac=frac)
    elif isinstance(example.answer, list):
        return _answer_match(pred.answer, example.answer, frac=frac)

    raise ValueError(f"Invalid answer type: {type(example.answer)}")


def answer_passage_match(example, pred, trace=None):
    """Return True if any passage in `pred.context` contains the answer(s).

    Strings are normalized (and passages also use DPR normalization internally).

    Args:
        example: `dspy.Example` object with field `answer` (str or list[str]).
        pred: `dspy.Prediction` object with field `context` (list[str]) containing passages.
        trace: Unused; reserved for compatibility.

    Returns:
        bool: True if any passage contains any reference answer; otherwise False.

    Examples:
        ```python
        import dspy

        example = dspy.Example(answer="Eiffel Tower")
        pred = dspy.Prediction(context=["The Eiffel Tower is in Paris.", "..."])

        answer_passage_match(example, pred)  # True
        ```
    """
    if isinstance(example.answer, str):
        return _passage_match(pred.context, [example.answer])
    elif isinstance(example.answer, list):
        return _passage_match(pred.context, example.answer)

    raise ValueError(f"Invalid answer type: {type(example.answer)}")



"""Usage tracking utilities for DSPy."""

from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Generator

from pydantic import BaseModel

from dspy.dsp.utils.settings import settings


class UsageTracker:
    """Tracks LM usage data within a context."""

    def __init__(self):
        # Map of LM name to list of usage entries. For example:
        # {
        #     "openai/gpt-4o-mini": [
        #         {"prompt_tokens": 100, "completion_tokens": 200},
        #         {"prompt_tokens": 300, "completion_tokens": 400},
        #     ],
        # }
        self.usage_data = defaultdict(list)

    def _flatten_usage_entry(self, usage_entry: dict[str, Any]) -> dict[str, Any]:
        result = {}
        for key, value in usage_entry.items():
            if isinstance(value, BaseModel):
                # Convert Pydantic models to dicts (e.g. PromptTokensDetailsWrapper).
                result[key] = value.model_dump()
            else:
                result[key] = value
        return result

    def _merge_usage_entries(
        self, usage_entry1: dict[str, Any] | None, usage_entry2: dict[str, Any] | None
    ) -> dict[str, Any]:
        if usage_entry1 is None or len(usage_entry1) == 0:
            return dict(usage_entry2)
        if usage_entry2 is None or len(usage_entry2) == 0:
            return dict(usage_entry1)

        result = dict(usage_entry2)
        for k, v in usage_entry1.items():
            current_v = result.get(k)
            if isinstance(v, dict) or isinstance(current_v, dict):
                result[k] = self._merge_usage_entries(current_v, v)
            elif current_v is not None or v is not None:
                result[k] = (current_v or 0) + (v or 0)
        return result

    def add_usage(self, lm: str, usage_entry: dict[str, Any]) -> None:
        """Add a usage entry to the tracker."""
        if len(usage_entry) > 0:
            self.usage_data[lm].append(self._flatten_usage_entry(usage_entry))

    def get_total_tokens(self) -> dict[str, dict[str, Any]]:
        """Calculate total tokens from all tracked usage."""
        total_usage_by_lm = {}
        for lm, usage_entries in self.usage_data.items():
            total_usage = {}
            for usage_entry in usage_entries:
                total_usage = self._merge_usage_entries(total_usage, usage_entry)
            total_usage_by_lm[lm] = total_usage
        return total_usage_by_lm


@contextmanager
def track_usage() -> Generator[UsageTracker, None, None]:
    """Context manager for tracking LM usage."""
    tracker = UsageTracker()

    with settings.context(usage_tracker=tracker):
        yield tracker



from dspy.evaluate import normalize_text
from dspy.primitives.prediction import Completions, Prediction


def default_normalize(s):
    return normalize_text(s) or None


def majority(prediction_or_completions, normalize=default_normalize, field=None):
    """
    Returns the most common completion for the target field (or the last field) in the signature.
    When normalize returns None, that completion is ignored.
    In case of a tie, earlier completion are prioritized.
    """

    assert any(isinstance(prediction_or_completions, t) for t in [Prediction, Completions, list])
    type(prediction_or_completions)

    # Get the completions
    if isinstance(prediction_or_completions, Prediction):
        completions = prediction_or_completions.completions
    else:
        completions = prediction_or_completions

    try:
        signature = completions.signature
    except Exception:
        signature = None

    if not field:
        if signature:
            field = list(signature.output_fields.keys())[-1]
        else:
            field = list(completions[0].keys())[-1]

    # Normalize
    normalize = normalize if normalize else lambda x: x
    normalized_values = [normalize(completion[field]) for completion in completions]
    normalized_values_ = [x for x in normalized_values if x is not None]

    # Count
    value_counts = {}
    for value in normalized_values_ or normalized_values:
        value_counts[value] = value_counts.get(value, 0) + 1

    majority_value = max(value_counts, key=value_counts.get)

    # Return the first completion with the majority value in the field
    for completion in completions:
        if normalize(completion[field]) == majority_value:
            break

    # if input_type == Prediction:
    return Prediction.from_completions([completion], signature=signature)



"""
Abstract interpreter interface for code execution environments.

This module defines the CodeInterpreter protocol that allows RLM and other
code-executing modules to work with different interpreter implementations:
- PythonInterpreter: Local Deno/Pyodide WASM interpreter
- MockInterpreter: Scriptable responses for testing
"""

from typing import Any, Callable, Protocol, runtime_checkable

# Types that can be used directly in Python function signatures for SUBMIT()
SIMPLE_TYPES = (str, int, float, bool, list, dict, type(None))


class CodeInterpreterError(RuntimeError):
    """Error raised during code interpretation.

    This exception covers two distinct failure modes:

    1. **Execution errors**: The sandbox ran user code that failed.
       - NameError, TypeError, ValueError, etc.
       - Tool call failures (unknown tool, tool raised exception)
       - These are normal user code errors.

    2. **Protocol errors**: Communication between host and sandbox failed.
       - Malformed JSON from sandbox
       - Sandbox process crashed or became unresponsive
       - Invalid JSON-RPC message structure
       - These may indicate a corrupted sandbox needing restart.

    The error message typically includes the original error type (e.g., "NameError: ...")
    which can help distinguish the failure mode.

    Note: SyntaxError is raised separately (not wrapped) for invalid Python syntax.
    """


class FinalOutput:
    """Returned by interpreter.execute() when SUBMIT() is called.

    This signals that the code execution loop should terminate and return
    the contained output to the caller.
    """

    def __init__(self, output: Any):
        self.output = output

    def __repr__(self) -> str:
        return f"FinalOutput({self.output!r})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FinalOutput):
            return NotImplemented
        return self.output == other.output


@runtime_checkable
class CodeInterpreter(Protocol):
    """Protocol for code execution environments (interpreters).

    Implementations must provide:
    - start(): Initialize the interpreter (optional, can be lazy)
    - execute(): Run code and return results
    - shutdown(): Clean up resources

    The interpreter maintains state across execute() calls within a session,
    allowing variables defined in one call to be used in subsequent calls.

    Lifecycle:
        1. Create instance (config only, no resources allocated)
        2. start() - Initialize interpreter (explicit) or let execute() do it (lazy)
        3. execute() - Run code (can be called many times)
        4. shutdown() - Release resources

    Example implementations:
        - LocalInterpreter: Deno/Pyodide WASM interpreter (local)
        - MockInterpreter: Scriptable responses for testing

    Pooling:
        For interpreter pooling, call start() to pre-warm instances, then
        distribute execute() calls across the pool.
    """

    @property
    def tools(self) -> dict[str, Callable[..., str]]:
        """Tools available for interpreter code to call.

        Tools are host-side functions that can be invoked from within the
        interpreter. Each tool accepts keyword arguments and returns a string.

        Implementations should accept tools via constructor and expose them
        through this property.
        """
        ...

    def start(self) -> None:
        """Initialize the interpreter and allocate resources.

        This method prepares the interpreter for code execution. It can be called
        explicitly to pre-warm the interpreter, or implementations may call it
        lazily on first execute().

        For pooling scenarios, call start() on multiple instances to have
        them ready before distributing work.

        Calling start() multiple times should be safe (idempotent).
        """
        ...

    def execute(
        self,
        code: str,
        variables: dict[str, Any] | None = None,
    ) -> Any:
        """Execute Python code and return the result.

        Args:
            code: Python code to execute
            variables: Variables to inject into the namespace before execution.
                      These are available as top-level variables in the code.

        Returns:
            One of:
            - FinalOutput: If SUBMIT() was called in code
            - str: Captured stdout from print() statements
            - list: Multiple output lines
            - None: If no output was produced

        Raises:
            CodeInterpreterError: On runtime errors (undefined vars, tool failures, etc.)
            SyntaxError: On invalid Python syntax

        Note:
            State persists across calls. Variables defined in one execute()
            call are available in subsequent calls until shutdown().

            If start() was not called, implementations should call it lazily.
        """
        ...

    def shutdown(self) -> None:
        """Release resources and terminate the interpreter session.

        After shutdown(), the interpreter should not be used again.
        A new instance should be created for a fresh session.
        """
        ...



import inspect
import logging
import random
from typing import Any, Callable

import dspy
from dspy.evaluate import Evaluate
from dspy.primitives.example import Example
from dspy.primitives.module import Module
from dspy.teleprompt.bootstrap_finetune import (
    BootstrapFinetune,
    all_predictors_have_lms,
    kill_lms,
    launch_lms,
    prepare_student,
    prepare_teacher,
)
from dspy.teleprompt.random_search import BootstrapFewShotWithRandomSearch
from dspy.teleprompt.teleprompt import Teleprompter
from dspy.teleprompt.utils import eval_candidate_program

logger = logging.getLogger(__name__)

# ANSI color codes for terminal output
YELLOW = "\033[93m"
GREEN = "\033[92m"
BLUE = "\033[94m"
BOLD = "\033[1m"
ENDC = "\033[0m"


class BetterTogether(Teleprompter):
    """A meta-optimizer that combines prompt and weight optimization in configurable sequences.

    BetterTogether is a meta-optimizer proposed in the paper
    [Fine-Tuning and Prompt Optimization: Two Great Steps that Work Better Together](https://arxiv.org/abs/2407.10930).
    It combines prompt optimization and weight optimization (fine-tuning) by applying them in a
    configurable sequence, allowing a student program to iteratively improve both its prompts and
    model parameters.

    The core insight is that prompt and weight optimization can complement each other: prompt
    optimization can potentially discover effective task decompositions and reasoning strategies,
    while weight optimization can specialize the model to execute these patterns more efficiently.
    Using these approaches together in sequences (e.g., prompt optimization then weight optimization)
    may allow each to build on the improvements made by the other. Empirically, this approach often
    outperforms either strategy alone, even with state-of-the-art optimizers. For example, a
    [Databricks case study](https://www.databricks.com/blog/building-state-art-enterprise-agents-90x-cheaper-automated-prompt-optimization)
    shows that combining BetterTogether with GEPA and fine-tuning outperforms either approach
    alone.

    The optimizer is initialized with a metric and custom optimizers. For example, you can combine
    GEPA for prompt optimization with BootstrapFinetune for weight optimization:
    ``BetterTogether(metric=metric, p=GEPA(...), w=BootstrapFinetune(...))``. The ``compile()``
    method takes a student program, trainset, and strategy string where strategy keys correspond
    to the optimizer names from initialization. It executes each optimizer in the specified sequence.
    When a validation set is provided, the best performing program is returned; otherwise, the
    latest program is returned. Note: Weight optimizers like BootstrapFinetune require student
    programs to have LMs explicitly set (not relying on global dspy.settings.lm), and BetterTogether
    mirrors this requirement for simplicity. Therefore we call ``set_lm`` before compiling.

        >>> from dspy.teleprompt import GEPA, BootstrapFinetune
        >>>
        >>> # Combine GEPA for prompt optimization with BootstrapFinetune for weight optimization
        >>> optimizer = BetterTogether(
        ...     metric=metric,
        ...     p=GEPA(metric=metric, auto="medium"),
        ...     w=BootstrapFinetune(metric=metric)
        ... )
        >>>
        >>> student.set_lm(lm)
        >>> compiled = optimizer.compile(
        ...     student,
        ...     trainset=trainset,
        ...     valset=valset,
        ...     strategy="p -> w"
        ... )

    You can pass optimizer-specific arguments to each optimizer's ``compile()`` method using
    ``optimizer_compile_args``. This allows you to customize each optimizer's behavior:

        >>> from dspy.teleprompt import MIPROv2
        >>>
        >>> # Use MIPROv2 for prompt optimization with custom parameters
        >>> optimizer = BetterTogether(
        ...     metric=metric,
        ...     p=MIPROv2(metric=metric),
        ...     w=BootstrapFinetune(metric=metric)
        ... )
        >>>
        >>> student.set_lm(lm)
        >>> compiled = optimizer.compile(
        ...     student,
        ...     trainset=trainset,
        ...     valset=valset,
        ...     strategy="p -> w",
        ...     optimizer_compile_args={
        ...         "p": {"num_trials": 10, "max_bootstrapped_demos": 8},  # Configure MIPROv2's compile arguments
        ...     }
        ... )

    Since BetterTogether is a meta-optimizer that can run arbitrary optimizers in sequence, any
    sequence of optimizers can be combined together. The optimizer names used in the strategy string
    correspond to the keyword arguments specified in the constructor. For example, different prompt
    optimizers can be alternated multiple times (though note this is just an illustration of
    BetterTogether's flexibility, not a recommended configuration):

        >>> from dspy.teleprompt import MIPROv2, GEPA
        >>>
        >>> # Chain two optimizers three times: MIPROv2 -> GEPA -> MIPROv2
        >>> optimizer = BetterTogether(
        ...     metric=metric,
        ...     mipro=MIPROv2(metric=metric, auto="light"),
        ...     gepa=GEPA(metric=metric, auto="light")
        ... )
        >>>
        >>> student.set_lm(lm)
        >>> compiled = optimizer.compile(
        ...     student,
        ...     trainset=trainset,
        ...     valset=valset,
        ...     strategy="mipro -> gepa -> mipro"
        ... )

    Note:
        Output Attributes: The returned program includes two additional attributes:
        ``candidate_programs`` and ``flag_compilation_error_occurred``. The ``candidate_programs``
        attribute is a list of dicts, each containing 'program', 'score', and 'strategy' (e.g.,
        '', 'p', 'p -> w', 'p -> w -> p'), sorted by descending score (similar to
        ``dspy.MIPROv2.candidate_programs``). If any optimizer step fails,
        ``flag_compilation_error_occurred`` is set to True and the best program found so far is
        returned.

        Model Lifecycle Management: BetterTogether automatically manages language model lifecycle
        (launching, killing, and relaunching after fine-tuning), which are no-ops for most
        API-based LMs. This is particularly important when using weight optimizers like
        BootstrapFinetune with local providers (e.g., dspy.LocalProvider), as it handles model
        initialization and cleanup between optimization steps.
    """

    STRAT_SEP = " -> "

    def __init__(
        self,
        metric: Callable,
        **optimizers: Teleprompter,
    ):
        """Initialize BetterTogether with a metric and custom optimizers.

        Args:
            metric: Evaluation metric function for scoring programs. Should accept
                ``(example, prediction, trace=None)`` and return a numeric score (higher is better).
                This metric is used to evaluate candidate programs during optimization and is passed
                to the default optimizers if no custom optimizers are provided.
            **optimizers: Custom optimizers as keyword arguments, where keys become the optimizer
                names used in the strategy string. For example, ``p=GEPA(...), w=BootstrapFinetune(...)``
                makes 'p' and 'w' available for use in strategies like ``"p -> w"``. If not provided,
                defaults to ``p=BootstrapFewShotWithRandomSearch(metric=metric)`` and
                ``w=BootstrapFinetune(metric=metric)``. Any DSPy Teleprompter can be used.

        Examples:
            >>> # Use custom optimizers
            >>> from dspy.teleprompt import GEPA, BootstrapFinetune
            >>> optimizer = BetterTogether(
            ...     metric=metric,
            ...     p=GEPA(metric=metric, auto="medium"),
            ...     w=BootstrapFinetune(metric=metric)
            ... )
            >>>
            >>> # Use default optimizers
            >>> optimizer = BetterTogether(metric=metric)
        """
        self.metric = metric

        if not optimizers:
            logger.info(
                "No optimizers provided. Using defaults: "
                "BootstrapFewShotWithRandomSearch (p) and BootstrapFinetune (w). "
                "You can use the letters p and w to specify the compile strategy. "
                "For example, to run weight optimization after prompt optimization, use strategy='p -> w'."
            )
            optimizers = {
                "p": BootstrapFewShotWithRandomSearch(metric=metric),
                "w": BootstrapFinetune(metric=metric),
            }
        for key, optimizer in optimizers.items():
            if not isinstance(optimizer, Teleprompter):
                raise TypeError(
                    f"Optimizer '{key}' must be a Teleprompter, "
                    f"got {type(optimizer).__name__}"
                )
        self.optimizers: dict[str, Teleprompter] = optimizers

    def compile(
        self,
        student: Module,
        *,
        trainset: list[Example],
        teacher: Module | list[Module] | None = None,
        valset: list[Example] | None = None,
        # often specified in init in other optimizers
        num_threads: int | None = None,
        max_errors: int | None = None,
        provide_traceback: bool | None = None,
        seed: int | None = None,
        # specific to BetterTogether
        valset_ratio: float = 0.1,
        shuffle_trainset_between_steps: bool = True,
        strategy: str = "p -> w -> p",
        optimizer_compile_args: dict[str, dict[str, Any]] | None = None,
    ) -> Module:
        """Compile and optimize a student program using a sequence of optimization strategies.

        Executes the optimizers specified in the strategy string sequentially, evaluating each
        intermediate result and returning the best performing program.

        Args:
            student: DSPy program to optimize. All predictors must have language models assigned.
                program.set_lm(lm) can be used to assign a language model to all modules of a 
                program.
            trainset: Training examples for optimization. Each optimizer receives the full trainset
                (or a shuffled version if ``shuffle_trainset_between_steps=True``).
            teacher: Optional teacher module(s) for bootstrapping. Can be a single module or list.
                Passed to optimizers.
            valset: Validation set for evaluating optimization steps. If not provided, a portion of
                trainset is held out (controlled by ``valset_ratio``). If both ``valset`` and
                ``valset_ratio`` are None/0, no validation occurs and the latest program is returned.
            num_threads: Number of parallel evaluation threads. Default is None, which means sequential evaluation.
            max_errors: Maximum errors to tolerate during evaluation. Defaults to
                ``dspy.settings.max_errors``.
            provide_traceback: Whether to show detailed tracebacks for evaluation errors.
            seed: Random seed for reproducibility. Controls trainset shuffling and evaluation sampling.
            valset_ratio: Fraction of trainset to hold out as validation (range [0, 1)). For example,
                0.1 holds out 10%. Set to 0 to skip validation. Default is 0.1.
            shuffle_trainset_between_steps: Whether to shuffle trainset before each optimization step.
                Helps prevent overfitting to example ordering. Default is True.
            strategy: Sequence of optimizers to apply, separated by ``" -> "``. Each element must be
                a key from the optimizers provided in ``__init__``. For example, ``"p -> w -> p"``
                applies prompt optimization, then weight optimization, then prompt optimization again.
                Default is ``"p -> w -> p"``.
            optimizer_compile_args: Optional dict mapping optimizer keys to their ``compile()``
                arguments. If trainset, valset, or teacher are provided in the dict for a specific
                optimizer, they override the defaults from BetterTogether's compile method. For example:
                ``{"p": {"num_trials": 10}, "w": {"trainset": custom_trainset}}``. This is useful to
                override the default compile arguments for specific optimizers. The ``student`` argument
                cannot be included in optimizer_compile_args; BetterTogether's compile method manages
                the student reference for all optimizers.

        Returns:
            Optimized student program with two additional attributes:

            - ``candidate_programs``: List of dicts with 'program', 'score', and 'strategy' keys,
              sorted by score (best first). Contains all evaluated programs including the baseline.
            - ``flag_compilation_error_occurred``: Boolean indicating if any optimization step failed.

        Raises:
            ValueError: If trainset is empty, valset_ratio not in [0, 1), strategy is empty or
                contains invalid optimizer keys, or optimizer_compile_args contains invalid arguments.
            TypeError: If optimizer_compile_args contains a 'student' key (not allowed).

        Examples:
            >>> optimizer = BetterTogether(
            ...     metric=metric,
            ...     p=GEPA(metric=metric),
            ...     w=BootstrapFinetune(metric=metric)
            ... )
            >>> student.set_lm(lm)
            >>> compiled = optimizer.compile(
            ...     student,
            ...     trainset=trainset,
            ...     valset=valset,
            ...     strategy="p -> w"
            ... )
            >>> print(f"Best score: {compiled.candidate_programs[0]['score']}")
        """
        logger.info(f"\n{BOLD}==> BETTERTOGETHER COMPILATION STARTED <=={ENDC}")
        logger.info(f"{BLUE}Strategy:{ENDC} {strategy}")
        logger.info(f"{BLUE}Trainset size:{ENDC} {len(trainset)}")
        logger.info(f"{BLUE}Validation ratio:{ENDC} {valset_ratio if valset is None else 'using provided valset'}")

        student, teacher = self._prepare_student_and_teacher(student, teacher)
        trainset, valset = self._prepare_trainset_and_valset(trainset, valset, valset_ratio)
        effective_max_errors = max_errors if max_errors is not None else dspy.settings.max_errors
        parsed_strategy = self._prepare_strategy(strategy)
        optimizer_compile_args = self._prepare_optimizer_compile_args(optimizer_compile_args, teacher)

        student = self._run_strategies(
            student,
            trainset,
            teacher,
            valset,
            num_threads,
            effective_max_errors,
            provide_traceback,
            seed,
            parsed_strategy,
            shuffle_trainset_between_steps,
            optimizer_compile_args,
        )

        logger.info(f"\n{BOLD}{GREEN}==> BETTERTOGETHER COMPILATION COMPLETE <=={ENDC}")
        logger.info(f"{GREEN}Best score achieved:{ENDC} {student.candidate_programs[0]['score']}")
        logger.info(f"{GREEN}Best strategy:{ENDC} {student.candidate_programs[0]['strategy'] or 'original (no optimization)'}")

        student._compiled = True
        return student

    def _prepare_student_and_teacher(
        self, student: Module, teacher: Module | list[Module] | None
    ) -> tuple[Module, list[Module] | None]:
        student = prepare_student(student)
        all_predictors_have_lms(student)

        if not teacher:
            return student, None
        teacher = [teacher] if not isinstance(teacher, list) else teacher
        teacher = [prepare_teacher(student, t) for t in teacher]

        return student, teacher

    def _prepare_trainset_and_valset(
        self, trainset: list[Example], valset: list[Example] | None, valset_ratio: float
    ) -> tuple[list[Example], list[Example] | None]:
        if not trainset:
            raise ValueError("trainset cannot be empty")

        if valset_ratio < 0 or valset_ratio >= 1:
            raise ValueError(f"valset_ratio must be in range [0, 1), got {valset_ratio}")

        trainset = trainset[:]  # shallow copy to avoid modifying the original

        if valset:
            logger.info(f"{BLUE}Using provided validation set ({len(valset)} examples). Ignoring valset_ratio.{ENDC}")
            return trainset, valset

        if valset_ratio == 0:
            logger.info(f"{YELLOW}No validation set provided and valset_ratio=0. No validation set created.{ENDC}")
            return trainset, None

        logger.info(f"{BLUE}Sampling {valset_ratio:.1%} of trainset as validation set.{ENDC}")
        num_val_examples = int(valset_ratio * len(trainset))
        valset = trainset[:num_val_examples]
        trainset = trainset[num_val_examples:]
        logger.info(f"{BLUE}Created validation set: {len(valset)} examples. Training set: {len(trainset)} examples.{ENDC}")

        return trainset, valset

    def _prepare_strategy(self, strategy: str) -> list[str]:
        if not strategy or not strategy.strip():
            raise ValueError("strategy cannot be empty")

        parsed_strategy = strategy.split(self.STRAT_SEP)

        invalid_steps = [s for s in parsed_strategy if s not in self.optimizers]
        if invalid_steps:
            raise ValueError(
                f"Strategy contains invalid optimizer keys: {invalid_steps}. "
                f"Valid keys are: {list(self.optimizers.keys())}"
            )

        return parsed_strategy

    def _prepare_optimizer_compile_args(
        self,
        optimizer_compile_args: dict[str, dict[str, Any]] | None,
        teacher: list[Module] | None,
    ) -> dict[str, dict[str, Any]]:
        logger.info(f"{BLUE}Validating optimizer compile arguments...{ENDC}")

        # Validate user-provided compile args
        if not optimizer_compile_args:
            return {}

        for optimizer_key, compile_args in optimizer_compile_args.items():
            if optimizer_key not in self.optimizers:
                raise ValueError(
                    f"Invalid optimizer key '{optimizer_key}'. "
                    f"Valid keys are: {list(self.optimizers.keys())}"
                )
            optimizer = self.optimizers[optimizer_key]
            self._validate_compile_args(optimizer, optimizer_key, compile_args)

            # Special checks
            if optimizer.__class__.__name__ == "GEPA":
                # GEPA accepts a teacher argument, but raises an error if it's set.
                if teacher is not None:
                    raise ValueError(
                        "GEPA does not accept a teacher argument. Please remove the teacher argument."
                    )

        return optimizer_compile_args

    def _validate_compile_args(
        self, optimizer: Teleprompter, optimizer_key: str, compile_args: dict[str, Any]
    ) -> None:
        if "student" in compile_args:
            raise ValueError(
                f"'student' is not allowed in optimizer_compile_args for optimizer '{optimizer_key}'. "
                "The same student is used throughout compilation."
            )

        valid_params = inspect.signature(optimizer.compile).parameters
        invalid_args = set(compile_args.keys()) - set(valid_params.keys())
        if invalid_args:
            raise ValueError(
                f"Invalid compile arguments for optimizer '{optimizer_key}': {sorted(invalid_args)}. "
                f"{optimizer.__class__.__name__}.compile() accepts: {list(valid_params.keys())}"
            )

    def _run_strategies(
        self,
        student: Module,
        trainset: list[Example],
        teacher: list[Module] | None,
        valset: list[Example] | None,
        num_threads: int | None,
        effective_max_errors: int | None,
        provide_traceback: bool | None,
        seed: int | None,
        parsed_strategy: list[str],
        shuffle_trainset_between_steps: bool,
        optimizer_args: dict[str, dict[str, Any]],
    ) -> Module:
        rng = random.Random(seed)
        candidate_programs = []
        flag_lms_launched = False
        flag_compilation_error_occurred = False

        # Evaluate original program
        logger.info(f"\n{BOLD}==> BASELINE EVALUATION <=={ENDC}")
        logger.info("Evaluating original program (no optimization applied)")

        launch_lms(student)
        flag_lms_launched = True

        score = self._evaluate_on_valset(student, valset, rng, num_threads, effective_max_errors, provide_traceback)
        self._add_candidate(candidate_programs, student, strategy="", score=score)
        logger.info(f"{YELLOW}Baseline score:{ENDC} {score}")

        # Apply each optimization step
        for ind, step_code in enumerate(parsed_strategy):
            current_strategy = self.STRAT_SEP.join(parsed_strategy[:ind + 1])
            optimizer = self.optimizers[step_code]

            logger.info(f"\n{BOLD}==> STEP {ind + 1}/{len(parsed_strategy)}: {optimizer.__class__.__name__.upper()} <=={ENDC}")
            logger.info(f"{BLUE}Current strategy:{ENDC} '{current_strategy}'")
            logger.info(f"{BLUE}Optimizer:{ENDC} {optimizer.__class__.__name__}")

            try:
                if shuffle_trainset_between_steps:
                    logger.info(f"{BLUE}Shuffling trainset...{ENDC}")
                    rng.shuffle(trainset)

                # Run optimizer, evaluate, and record results
                compile_args = optimizer_args.get(step_code, {})
                student, score, is_new_best, lms_relaunched = self._run_and_evaluate_step(
                    optimizer, student, teacher, trainset, valset, compile_args,
                    candidate_programs, current_strategy, rng,
                    num_threads, effective_max_errors, provide_traceback
                )

                if lms_relaunched:
                    flag_lms_launched = True

                # Log score
                if is_new_best:
                    logger.info(f"{GREEN}New best score!{ENDC} {score} (strategy: '{current_strategy}')")
                else:
                    logger.info(f"{YELLOW}Score after optimization:{ENDC} {score}")

            except Exception as e:
                flag_compilation_error_occurred = True
                logger.error(
                    f"{YELLOW}Step {ind + 1}/{len(parsed_strategy)} failed with error: {type(e).__name__}: {e}{ENDC}"
                )
                logger.error(
                    f"{YELLOW}Stopping optimization early. Returning best program found so far from {len(candidate_programs)} candidate(s).{ENDC}"
                )
                logger.error(f"{YELLOW}Traceback:{ENDC}", exc_info=True)
                break

        # Cleanup and finalize
        if flag_lms_launched:
            kill_lms(student)

        # Sort candidates by score (best first), with earlier programs winning ties
        candidate_programs_with_idx = [(i, cp) for i, cp in enumerate(candidate_programs)]
        candidate_programs_with_idx.sort(
            key=lambda x: (x[1]["score"] if x[1]["score"] is not None else float("-inf"), -x[0]),
            reverse=True
        )
        candidate_programs = [cp for _, cp in candidate_programs_with_idx]

        # Select best program based on valset availability
        if valset is None or len(valset) == 0:
            # No valset: return the latest program (last in original order)
            best_program = candidate_programs_with_idx[-1][1]
        else:
            # Valset provided: return highest score (first after sorting)
            best_program = candidate_programs[0]

        # Attach sorted candidate programs and error flag to the best program
        best_student = best_program["program"]
        best_student.candidate_programs = candidate_programs
        best_student.flag_compilation_error_occurred = flag_compilation_error_occurred

        logger.info(f"\n{BOLD}==> OPTIMIZATION SUMMARY <=={ENDC}")
        logger.info(f"{GREEN}Best score:{ENDC} {best_program['score']}")
        strategy_display = best_program["strategy"] if best_program["strategy"] else "original (no optimization)"
        logger.info(f"{GREEN}Best strategy:{ENDC} {strategy_display}")
        logger.info(f"{BLUE}Total candidates evaluated:{ENDC} {len(candidate_programs)}")

        return best_student

    def _run_and_evaluate_step(
        self,
        optimizer: Teleprompter,
        student: Module,
        teacher: list[Module] | None,
        trainset: list[Example],
        valset: list[Example] | None,
        compile_args: dict[str, Any],
        candidate_programs: list,
        current_strategy: str,
        rng: random.Random,
        num_threads: int | None,
        effective_max_errors: int | None,
        provide_traceback: bool | None,
    ) -> tuple[Module, float | None, bool, bool]:
        """Run optimizer, evaluate result, and record candidate program.

        Returns:
            student: Optimized student program
            score: Validation score (None if no valset)
            is_new_best: Whether this score is the best so far
            lms_relaunched: Whether LMs were relaunched (for flag tracking)
        """
        # Save LMs before optimization
        pred_lms_before = [pred.lm for pred in student.predictors()]

        # Run optimizer
        student._compiled = False
        logger.info(f"{BLUE}Running {optimizer.__class__.__name__} with {len(trainset)} training examples...{ENDC}")

        # Build compile args with standard parameters, filtering to only what the optimizer accepts
        potential_args = {"trainset": trainset, "teacher": teacher, "valset": valset, **compile_args}
        sig = inspect.signature(optimizer.compile)
        accepted_params = set(sig.parameters.keys())
        filtered_compile_args = {k: v for k, v in potential_args.items() if k in accepted_params}

        student = optimizer.compile(student, **filtered_compile_args)

        # Restore LMs if optimizer incorrectly reset them
        # TODO: Some optimizers like BootstrapFewShotWithRandomSearch reset predictor LMs during compilation,
        # which breaks weight optimizers as they require each predictor to have an LM. We should ensure that
        # all optimizers respect the assigned LMs of programs and do not override them. Until then, we restore
        # the original LMs here.
        if not all_predictors_have_lms(student):
            logger.warning(
                f"{YELLOW}Warning: {optimizer.__class__.__name__} incorrectly reset predictor LMs. "
                f"Restoring to original LMs.{ENDC}"
            )
            for pred, lm in zip(student.predictors(), pred_lms_before, strict=False):
                pred.lm = lm

        # Relaunch LMs if models changed (e.g., after fine-tuning)
        # Weight optimizers create new model instances and kill the old ones before returning.
        # We detect this via model name changes and relaunch the new models.
        # Note: launch() and kill() are no-ops for most API-based LMs; these mainly affect local
        # LMs that need launch or clean up routines.
        lms_relaunched = False
        if self._models_changed(student, pred_lms_before):
            launch_lms(student)
            lms_relaunched = True

        # Evaluate optimized program
        score = self._evaluate_on_valset(student, valset, rng, num_threads, effective_max_errors, provide_traceback)
        self._add_candidate(candidate_programs, student, current_strategy, score)

        # Check if this is the best score so far
        valid_scores = [cp["score"] for cp in candidate_programs if cp["score"] is not None]
        best_score_so_far = max(valid_scores) if valid_scores else float("-inf")
        is_new_best = score is not None and score >= best_score_so_far

        return student, score, is_new_best, lms_relaunched

    def _models_changed(self, student: Module, pred_lms_before: list) -> bool:
        """Check if model names changed after optimization (e.g., fine-tuning)."""
        pred_lms_after = [pred.lm for pred in student.predictors()]
        model_names_before = [lm.model if lm else None for lm in pred_lms_before]
        model_names_after = [lm.model if lm else None for lm in pred_lms_after]
        return model_names_before != model_names_after

    def _add_candidate(
        self,
        candidate_programs: list,
        student: Module,
        strategy: str,
        score: float,
    ) -> None:
        """Add a candidate program to the list."""
        candidate_programs.append({
            "score": score,
            "program": student.deepcopy(),
            "strategy": strategy,
        })

    def _evaluate_on_valset(
        self,
        program: Module,
        valset: list[Example] | None,
        rng: random.Random,
        num_threads: int | None,
        effective_max_errors: int | None,
        provide_traceback: bool | None,
    ) -> float | None:
        if valset is None or len(valset) == 0:
            logger.info(f"{YELLOW}No validation set provided. Skipping evaluation.{ENDC}")
            return None

        logger.info(f"{BLUE}Evaluating on {len(valset)} validation examples...{ENDC}")
        evaluate = Evaluate(
            devset=valset,
            metric=self.metric,
            num_threads=num_threads,
            max_errors=effective_max_errors,
            display_table=False,
            display_progress=True,
            provide_traceback=provide_traceback,
        )
        eval_result = eval_candidate_program(len(valset), valset, program, evaluate, rng)
        return eval_result.score
